import logging
import numpy as np
from IPython.display import clear_output
import os
import time

from timeout_decorator import timeout
from typing import Optional

from autoconf import conf
from autoconf import cached_property

from autofit import exc

from autofit.text import text_util


from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.analysis import Analysis



def get_timeout_seconds():

    try:
        return conf.instance["general"]["test"]["lh_timeout_seconds"]
    except KeyError:
        pass

logger = logging.getLogger(__name__)
timeout_seconds = get_timeout_seconds()

class Fitness:
    def __init__(
        self,
        model : AbstractPriorModel,
        analysis : Analysis,
        paths : Optional[AbstractPaths] = None,
        fom_is_log_likelihood: bool = True,
        resample_figure_of_merit: float = None,
        convert_to_chi_squared: bool = False,
        store_history: bool = False,
        use_jax_vmap : bool = False,
        batch_size : Optional[int] = None,
        iterations_per_quick_update: Optional[int] = None,
        xp=np,
    ):
        """
        Interfaces with any non-linear search to fit the model to the data and return a log likelihood via
        the analysis.

        The interface of a non-linear search and fitness function is summarised as follows:

        1) The non-linear search samples a new set of model parameters, which are passed to the fitness
        function's `__call__` method.

        2) The list of parameter values are mapped to an instance of the model.

        3) The instance is passed to the analysis class's log likelihood function, which fits the model to the
        data and returns the log likelihood.

        4) A final figure-of-merit is computed and returned to the non-linear search, which is either the log
        likelihood or log posterior (e.g. adding the log prior to the log likelihood).

        Certain searches (commonly nested samplers) require the parameters to be mapped from unit values to physical
        values, which is performed internally by the fitness object in step 2.

        Certain searches require the returned figure of merit to be a log posterior (often MCMC methods) whereas
        others require it to be a log likelihood (often nested samples which account for priors internally) in step 4.
        Which values is returned by the `fom_is_log_likelihood` bool.

        Some searches require a chi-squared value (which they minimized), given by the log likelihood multiplied
        by -2.0. This is returned by the fitness if the `convert_to_chi_squared` bool is `True`.

        If a model-fit raises an exception or returns a `np.nan`, a `resample_figure_of_merit` value is returned
        instead. The appropriate value depends on the search, but is typically either `None`, `-np.inf` or `1.0e99`.
        All values indicate to the non-linear search that the model-fit should be resampled or ignored.

        Many searches do not store the history of the parameters and log likelihood values, often to save
        memory on large model-fits. However, this can be useful, for example to plot the results of a model-fit
        versus iteration number. If the `store_history` bool is `True`, the parameters and log likelihoods are stored
        in the `parameters_history_list` and `figure_of_merit_history_list` attribute of the fitness object.

        Parameters
        ----------
        analysis
            An object that encapsulates the data and a log likelihood function which fits the model to the data
            via the non-linear search.
        model
            The model that is fitted to the data, which is used by the non-linear search to create instances of
            the model that are fitted to the data via the log likelihood function.
        paths
            The paths of the search, which if the search is being resumed from an old run is used to check that
            the likelihood function has not changed from the previous run.
        fom_is_log_likelihood
            If `True`, the figure of merit returned by the fitness function is the log likelihood. If `False`, the
            figure of merit is the log posterior.
        resample_figure_of_merit
            The figure of merit returned if the model-fit raises an exception or returns a `np.nan`.
        convert_to_chi_squared
            If `True`, the figure of merit returned is the log likelihood multiplied by -2.0, such that it is a
            chi-squared value that is minimized.
        store_history
            If `True`, the parameters and log likelihood values of every model-fit are stored in lists.
        """

        self.analysis = analysis
        self.model = model
        self.paths = paths
        self.fom_is_log_likelihood = fom_is_log_likelihood
        self.resample_figure_of_merit = resample_figure_of_merit or -xp.inf
        self.convert_to_chi_squared = convert_to_chi_squared
        self.store_history = store_history

        self.parameters_history_list = []
        self.log_likelihood_history_list = []

        self.use_jax_vmap = use_jax_vmap

        self._call = self.call

        if self.use_jax_vmap:
            self._call = self._vmap

        self.batch_size = batch_size
        self.iterations_per_quick_update = iterations_per_quick_update
        self.quick_update_max_lh_parameters = None
        self.quick_update_max_lh = -xp.inf
        self.quick_update_count = 0

        if self.paths is not None:
            self.check_log_likelihood(fitness=self)

    @property
    def _xp(self):
        if self.analysis._use_jax:
            import jax.numpy as jnp
            return jnp
        return np

    def call(self, parameters):
        """
        A private method that calls the fitness function with the given parameters and additional keyword arguments.
        This method is intended for internal use only.

        Parameters
        ----------
        parameters
            The parameters (typically a list) chosen by a non-linear search, which are mapped to an instance of the
            model via its priors and fitted to the data.
        kwargs
            Additional key-word arguments that may be necessary for specific non-linear searches.

        Returns
        -------
        The figure of merit returned to the non-linear search, which is either the log likelihood or log posterior.
        """
        # Get instance from model
        instance = self.model.instance_from_vector(vector=parameters, xp=self._xp)

        if self._xp.__name__.startswith("jax"):

            # Evaluate log likelihood (must be side-effect free and exception-free)
            log_likelihood = self.analysis.log_likelihood_function(instance=instance)

        else:

            try:
                log_likelihood = self.analysis.log_likelihood_function(instance=instance)
            except exc.FitException:
                return self.resample_figure_of_merit

        # Penalize NaNs in the log-likelihood
        log_likelihood = self._xp.where(self._xp.isnan(log_likelihood), self.resample_figure_of_merit, log_likelihood)
        log_likelihood = self._xp.where(self._xp.isinf(log_likelihood), self.resample_figure_of_merit, log_likelihood)

        # Determine final figure of merit
        if self.fom_is_log_likelihood:
            figure_of_merit = log_likelihood
        else:
            # Ensure prior list is compatible with JAX (must return a JAX array, not list)
            log_prior_array = self._xp.array(self.model.log_prior_list_from_vector(vector=parameters, xp=self._xp))
            figure_of_merit = log_likelihood + self._xp.sum(log_prior_array)

        # Convert to chi-squared scale if requested
        if self.convert_to_chi_squared:
            figure_of_merit *= -2.0

        return figure_of_merit

    def call_wrap(self, parameters):
        """
        Wrapper around a JAX-jitted likelihood function that optionally stores
        the history of evaluated parameters and likelihood values.

        Depending on whether the figure of merit
        (FoM) is defined as a log-likelihood (`self.fom_is_log_likelihood`), it
        either uses the FoM directly or subtracts the summed log-prior to obtain
        the log-likelihood.

        If `self.store_history` is True, both the input parameters and the
        corresponding log-likelihood are appended to internal history lists
        (`self.parameters_history_list`, `self.log_likelihood_history_list`).

        Parameters
        ----------
        parameters
            A vector of model parameters to evaluate.

        Returns
        -------
        float
            The computed figure of merit for the input parameters. This is either
            the log-likelihood itself or another objective function value,
            depending on configuration.
        """

        if self.use_jax_vmap:
            if len(np.array(parameters).shape) == 1:
                parameters = np.array(parameters)[None, :]

        figure_of_merit = self._call(parameters)

        if self.convert_to_chi_squared:
            log_likelihood = -0.5 * figure_of_merit
        else:
            log_likelihood = figure_of_merit

        if not self.fom_is_log_likelihood:
            log_prior_list = np.array(self.model.log_prior_list_from_vector(vector=parameters, xp=np))
            log_likelihood -= np.sum(log_prior_list)

        self.manage_quick_update(parameters=parameters, log_likelihood=log_likelihood)

        if self.store_history:

            self.parameters_history_list.append(np.array(parameters))
            self.log_likelihood_history_list.append(np.array(log_likelihood))

        return figure_of_merit

    def manage_quick_update(self, parameters, log_likelihood):
        """
        Manage quick updates during the non-linear search.

        A "quick update" is a lightweight visualization of the current best-fit
        (maximum likelihood) model parameters. This provides fast feedback on the
        progress of the fit without waiting for the full analysis to complete.

        It does not require leaving the active non-linear search, and is
        therefore faster than the full analysis visualization.

        Workflow:
        ----------
        1. Track the number of likelihood evaluations since the last quick update.
        2. Identify the maximum log-likelihood from the current batch of evaluations.
           - If `log_likelihood` is an array (batched evaluations), find the best
             index with `argmax`.
           - If it’s just a scalar (single evaluation), treat it as one update.
        3. If a new maximum likelihood is found, update:
           - `self.quick_update_max_lh` (best log-likelihood value so far).
           - `self.quick_update_max_lh_parameters` (corresponding parameter vector).
        4. Once the number of evaluations exceeds
           `self.iterations_per_quick_update`, generate a quick visualization of
           the current max-likelihood model via
           `self.analysis.perform_quick_update()`.

        Parameters
        ----------
        parameters : array-like
            The parameter vectors evaluated in this batch. Shape is typically
            (n_batch, n_param).
        log_likelihood : float or array-like
            The corresponding log-likelihood(s). If batched, must have shape
            (n_batch,).

        Notes
        -----
        - Quick updates are optional and controlled by
          `self.iterations_per_quick_update`.
        - If the `analysis` class does not implement
          `perform_quick_update`, the update is silently skipped.
        - This mechanism is intended for fast, coarse visualization only,
          not detailed science-quality outputs.
        """

        if self.iterations_per_quick_update is None:
            return

        try:

            best_idx = self._xp.argmax(log_likelihood)
            best_log_likelihood = log_likelihood[best_idx]
            best_parameters = parameters[best_idx]
            total_updates = log_likelihood.shape[0]

        except (AttributeError, IndexError, TypeError):

            best_log_likelihood = log_likelihood
            best_parameters = parameters
            total_updates = 1

        if best_log_likelihood > self.quick_update_max_lh:
            self.quick_update_max_lh = best_log_likelihood
            self.quick_update_max_lh_parameters = best_parameters

        self.quick_update_count += total_updates

        if self.quick_update_count >= self.iterations_per_quick_update:

            clear_output(wait=True)

            start_time = time.time()

            logger.info("Performing quick update of maximum log likelihood fit image and model.results")

            instance = self.model.instance_from_vector(vector=self.quick_update_max_lh_parameters, xp=self._xp)

            try:
                self.analysis.perform_quick_update(self.paths, instance)
            except NotImplementedError:
                pass

            result_info = text_util.result_max_lh_info_from(
                max_log_likelihood_sample=self.quick_update_max_lh_parameters.tolist(),
                max_log_likelihood=self.quick_update_max_lh,
                model=self.model,
            )
            result_info = "\n".join(result_info)

            logger.info(result_info)
            self.paths.output_model_results(result_info=result_info)

            self.quick_update_count = 0

            logger.info(f"Quick update complete in {time.time() - start_time} seconds.")

    @timeout(timeout_seconds)
    def __call__(self, parameters, *kwargs):
        """
        Interfaces with any non-linear in order to fit a model to the data and return a log likelihood via
        an `Analysis` class.

        The interface is described in full in the `__init__` docstring above.

        Parameters
        ----------
        parameters
            The parameters (typically a list) chosen by a non-linear search, which are mapped to an instance of the
            model via its priors and fitted to the data.
        kwargs
            Addition key-word arguments that may be necessary for specific non-linear searches.

        Returns
        -------
        The figure of merit returned to the non-linear search, which is either the log likelihood or log posterior.
        """
        return self.call_wrap(parameters)

    # def __getstate__(self):
    #     state = self.__dict__.copy()
    #     # Remove non-pickleable attributes
    #     state.pop('_call', None)
    #     state.pop('_grad', None)
    #     return state
    #
    # def __setstate__(self, state):
    #     self.__dict__.update(state)

    @cached_property
    def _vmap(self):
        """
        Vectorized and JIT-compiled likelihood function.

        This wraps the base likelihood function (`self.call`) with both
        `jax.jit` and `jax.vmap`, producing a function that can evaluate
        batches of parameter vectors efficiently in parallel. The first
        call incurs compilation time, but subsequent calls are highly
        optimized.

        Because this is a `cached_property`, the compiled function is stored
        after its first creation, avoiding repeated JIT compilation overhead.
        """
        import jax
        start = time.time()
        logger.info("JAX: Applying vmap and jit to likelihood function -- may take a few seconds.")
        func = jax.vmap(jax.jit(self.call))
        logger.info(f"JAX: vmap and jit applied in {time.time() - start} seconds.")
        return func

    @cached_property
    def _jit(self):
        """
        JIT-compiled likelihood function.

        This wraps the base likelihood function (`self.call`) with `jax.jit`,
        producing a compiled version optimized for repeated evaluation on a
        single set of parameters. The first call triggers compilation, while
        later calls benefit from the compiled execution.

        As a `cached_property`, the compiled function is cached after its
        first use, so JIT compilation only occurs once.
        """
        import jax
        start = time.time()
        logger.info("JAX: Applying jit to likelihood function -- may take a few seconds.")
        func = jax.jit(self.call)
        logger.info(f"JAX: jit applied in {time.time() - start} seconds.")
        return func

    @cached_property
    def _grad(self):
        """
        Gradient of the JIT-compiled likelihood function.

        This wraps the JIT-compiled likelihood function (`self._call`) with
        `jax.grad`, returning a function that computes gradients of the
        likelihood with respect to its input parameters. Useful for gradient-
        based optimization and inference methods.

        Since this is a `cached_property`, the gradient function is compiled
        and cached on first access, ensuring that expensive setup is done
        only once.
        """
        import jax
        start = time.time()
        logger.info("JAX: Applying grad to likelihood function -- may take a few seconds.")
        func = jax.grad(self.call)
        logger.info(f"JAX: grad applied in {time.time() - start} seconds.")
        return func

    def grad(self, *args, **kwargs):
        return self._grad(*args, **kwargs)

    def check_log_likelihood(self, fitness):
        """
        Changes to the PyAutoGalaxy source code may inadvertantly change the numerics of how a log likelihood is
        computed. Equally, one may set off a model-fit that resumes from previous results, but change the settings of
        the pixelization or inversion in a way that changes the log likelihood function.

        This function performs an optional sanity check, which raises an exception if the log likelihood calculation
        changes, to ensure a model-fit is not resumed with a different likelihood calculation to the previous run.

        If the model-fit has not been performed before (e.g. it is not a resume) this function outputs
        the `figure_of_merit` (e.g. the log likelihood) of the maximum log likelihood model at the end of the model-fit.

        If the model-fit is a resume, it loads this `figure_of_merit` and compares it against a new value computed for
        the resumed run (again using the maximum log likelihood model inferred). If the two likelihoods do not agree
        and therefore the log likelihood function has changed, an exception is raised and the code execution terminated.

        Parameters
        ----------
        paths
            certain searches the non-linear search outputs are stored,
            visualization, and pickled objects used by the database and aggregator.
        result
            The result containing the maximum log likelihood fit of the model.
        """
        import numpy as np

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return

        if not conf.instance["general"]["test"]["check_likelihood_function"]:
            return

        try:
            samples_summary = self.paths.load_samples_summary()
        except FileNotFoundError:
            return

        try:
            max_log_likelihood_sample = samples_summary.max_log_likelihood_sample
        except AttributeError:
            return
        log_likelihood_old = samples_summary.max_log_likelihood_sample.log_likelihood

        parameters = max_log_likelihood_sample.parameter_lists_for_model(model=self.model)

        log_likelihood_new = fitness(parameters=parameters)

        if not np.isclose(log_likelihood_old, log_likelihood_new):
            raise exc.SearchException(
                f"""
                Figure of merit sanity check failed. 

                This means that the existing results of a model fit used a different
                likelihood function compared to the one implemented now.
                Old Figure of Merit = {log_likelihood_old}
                New Figure of Merit = {log_likelihood_new}
                """
            )


class LinearRelationship:
    def __init__(self, m: float, c: float):
        """
        Describes a linear relationship between x and y, y = mx + c

        Parameters
        ----------
        m
            The gradient of the relationship
        c
            The y-intercept of the relationship
        """
        self.m = m
        self.c = c

    def __call__(self, x: float) -> float:
        """
        Calculate the value of y for a given value of x
        """
        return self.m * x + self.c

    def __str__(self):
        return f"y = {self.m}x + {self.c}"

    def __repr__(self):
        return str(self)



import itertools
import random
from abc import ABC, abstractmethod
from copy import copy
from typing import Union, Tuple, Optional, Dict

from autoconf import conf

from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior.constant import Constant
from autofit.mapper.prior.deferred import DeferredArgument
from autofit.mapper.variable import Variable

epsilon = 1e-14


class Prior(Variable, ABC, ArithmeticMixin):
    __database_args__ = ("id_")

    _ids = itertools.count()

    def __init__(self, message, id_=None):
        """
        An object used to mappers a unit value to an attribute value for a specific
        class attribute.

        Parameters
        ----------
        message

        """
        super().__init__(id_=id_)
        self.message = message
        message.id_ = self.id

        self.width_modifier = None

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Create a prior from a flattened PyTree

        Parameters
        ----------
        aux_data
            Auxiliary information that remains unchanged including
            the keys of the dict
        children
            Child objects subject to change

        Returns
        -------
        An instance of this class
        """
        return cls(*children)

    def unit_value_for(self, physical_value: float) -> float:
        """
        Compute the unit value between 0 and 1 for the physical value.
        """
        return self.message.cdf(physical_value)

    def with_message(self, message):
        new = copy(self)
        new.message = message
        return new

    def new(self):
        """
        Returns a copy of this prior with a new id assigned making it distinct
        """
        new = copy(self)
        new.id = next(self._ids)
        return new

    @abstractmethod
    def with_limits(self, lower_limit: float, upper_limit: float) -> "Prior":
        """
        Create a new instance of the same prior class with the passed limits.
        """

    @property
    def factor(self):
        """
        A callable PDF used as a factor in factor graphs
        """
        return self.message.factor

    @staticmethod
    def for_class_and_attribute_name(cls, attribute_name):
        prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name]
        )
        return Prior.from_dict(prior_dict)

    def random(
        self,
        lower_limit : float = 0.0,
        upper_limit : float = 1.0
    ) -> float:
        """
        A random value sampled from this prior
        """
        return self.value_for(
            random.uniform(
                lower_limit,
                upper_limit,
            )
        )

    def value_for(self, unit: float) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        A physical value, mapped from the unit value accoridng to the prior.
        """
        return self.message.value_for(unit)

    def instance_for_arguments(
        self,
        arguments,
        ignore_assertions=False,
    ):
        _ = ignore_assertions
        return arguments[self]

    def project(self, samples, weights):
        result = copy(self)
        result.message = self.message.project(
            samples=samples,
            log_weight_list=weights,
            id_=self.id,
        )
        return result

    def __getattr__(self, item):
        if item in ("__setstate__", "__getstate__"):
            raise AttributeError(item)
        return getattr(self.message, item)

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<{} id={}>".format(
            self.__class__.__name__, self.id,
        )

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"{self.__class__.__name__} [{self.id}], {self.parameter_string}"

    @property
    @abstractmethod
    def parameter_string(self) -> str:
        pass

    def __float__(self):
        return self.value_for(0.5)

    @classmethod
    def from_dict(
        cls,
        prior_dict: dict,
        reference: Optional[Dict[str, str]] = None,
        loaded_ids: Optional[Dict[int, "Prior"]] = None,
    ) -> Union["Prior", DeferredArgument]:
        """
        Returns a prior from a JSON representation.

        Parameters
        ----------
        prior_dict : dict
            A dictionary representation of a prior including a type (e.g. Uniform) and all constructor arguments.
        reference
            A dictionary mapping prior ids to the priors they reference.
        loaded_ids
            A dictionary mapping prior ids to the priors they have loaded.

        Returns
        -------
        An instance of a child of this class.
        """
        loaded_ids = {} if loaded_ids is None else loaded_ids
        id_ = prior_dict.get("id")
        try:
            return loaded_ids[id_]
        except KeyError:
            pass
        if prior_dict["type"] == "Deferred":
            return DeferredArgument()

        from .uniform import UniformPrior
        from .log_uniform import LogUniformPrior
        from .gaussian import GaussianPrior
        from .log_gaussian import LogGaussianPrior
        from .truncated_gaussian import TruncatedGaussianPrior

        prior_type_dict = {
            "Uniform": UniformPrior,
            "LogUniform": LogUniformPrior,
            "Gaussian": GaussianPrior,
            "LogGaussian": LogGaussianPrior,
            "TruncatedGaussian" : TruncatedGaussianPrior,
            "Constant": Constant,
        }

        # noinspection PyProtectedMember
        prior = prior_type_dict[prior_dict["type"]](
            **{
                key: value
                for key, value in prior_dict.items()
                if key not in ("type", "width_modifier", "limits", "id")
            },
        )
        if id_ is not None:
            loaded_ids[id_] = prior

        return prior

    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = {
            "type": self.name_of_class(),
            "id": self.id,
        }
        return prior_dict

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("Prior", "")

    @property
    def limits(self) -> Tuple[float, float]:
        return (float("-inf"), float("inf"))

    def gaussian_prior_model_for_arguments(self, arguments):
        return arguments[self]



from abc import ABC
from typing import Optional
import warnings

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.search.abstract_search import NonLinearSearch
from autofit.non_linear.initializer import (
    InitializerPrior,
    AbstractInitializer,
    InitializerParamBounds,
)
from autofit.non_linear.samples import SamplesNest
from autofit.non_linear.plot.nest_plotters import NestPlotter
from autofit.non_linear.plot.output import Output


class AbstractNest(NonLinearSearch, ABC):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        iterations_per_quick_update: Optional[int] = None,
        iterations_per_full_update: Optional[int] = None,
        number_of_cores: Optional[int] = None,
        session: Optional[sa.orm.Session] = None,
        initializer: Optional[AbstractInitializer] = None,
        **kwargs
    ):
        """
        Abstract class of a nested sampling `NonLinearSearch` (e.g. MultiNest, Dynesty).

        **PyAutoFit** allows a nested sampler to automatically terminate when the acceptance ratio falls below an input
        threshold value. When this occurs, all samples are accepted using the current maximum log likelihood value,
        irrespective of how well the model actually fits the data.

        This feature should be used for non-linear searches where the nested sampler gets 'stuck', for example because
        the log likelihood function is stochastic or varies rapidly over small scales in parameter space. The results of
        samples using this feature are not realiable (given the log likelihood is being manipulated to end the run), but
        they are still valid results for linking priors to a new search and non-linear search.

        Parameters
        ----------
        session
            An SQLAlchemy session instance so the results of the model-fit are written to an SQLite database.
        """
        if isinstance(initializer, InitializerParamBounds):
            raise ValueError(
                "InitializerParamBounds cannot be used for nested sampling"
            )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer or InitializerPrior(),
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

    @property
    def config_type(self):
        return conf.instance["non_linear"]["nest"]

    @property
    def samples_cls(self):
        return SamplesNest

    @property
    def plotter_cls(self):
        return NestPlotter

    def plot_results(self, samples):

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["nest"][name]

        plotter = self.plotter_cls(
            samples=samples,
            output=Output(path=self.paths.image_path / "search", format="png"),
        )
        if should_plot("corner_anesthetic"):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                plotter.corner_anesthetic()


from typing import Tuple, Dict, Optional, Union

from autoconf.dictable import from_dict
from .abstract import AbstractPriorModel
from autofit.mapper.prior.abstract import Prior
import numpy as np


class Array(AbstractPriorModel):
    def __init__(
        self,
        shape: Tuple[int, ...],
        prior: Optional[Prior] = None,
    ):
        """
        An array of priors.

        Parameters
        ----------
        shape : (int, int)
            The shape of the array.
        prior : Prior
            The prior of every entry in the array.
        """
        super().__init__()
        self.shape = shape
        self.indices = list(np.ndindex(*shape))

        if prior is not None:
            for index in self.indices:
                self[index] = prior.new()

    @staticmethod
    def _make_key(index: Tuple[int, ...]) -> str:
        """
        Make a key for the prior.

        This is so an index (e.g. (1, 2)) can be used to access a
        prior (e.g. prior_1_2).

        Parameters
        ----------
        index
            The index of an element in an array.

        Returns
        -------
        The attribute name for the prior.
        """
        if isinstance(index, int):
            suffix = str(index)
        else:
            suffix = "_".join(map(str, index))
        return f"prior_{suffix}"

    def _instance_for_arguments(
        self,
        arguments: Dict[Prior, float],
        ignore_assertions: bool = False,
        xp=np,
    ) -> np.ndarray:
        """
        Create an array where the prior at each index is replaced with the
        a concrete value.

        Parameters
        ----------
        arguments
            The arguments to replace the priors with.
        ignore_assertions
            Whether to ignore assertions in the priors.

        Returns
        -------
        The array with the priors replaced.
        """
        make_array = True

        for index in self.indices:
            value = self[index]
            try:
                value = value.instance_for_arguments(
                    arguments,
                    ignore_assertions,
                )
            except AttributeError:
                pass

            if make_array:
                if isinstance(value, np.ndarray) or isinstance(value, np.float64):
                    array = np.zeros(self.shape)
                    make_array = False
                else:
                    import jax.numpy as jnp
                    array = jnp.zeros(self.shape)
                    make_array = False

            if isinstance(value, np.ndarray) or isinstance(value, np.float64):
                array[index] = value
            else:
                array = array.at[index].set(value)

        return array

    def __setitem__(
        self,
        index: Union[int, Tuple[int, ...]],
        value: Union[float, Prior],
    ):
        """
        Set the value at an index.

        Parameters
        ----------
        index
            The index of the prior.
        value
            The new value.
        """
        setattr(
            self,
            self._make_key(index),
            value,
        )

    def __getitem__(
        self,
        index: Union[int, Tuple[int, ...]],
    ) -> Union[float, Prior]:
        """
        Get the value at an index.

        Parameters
        ----------
        index
            The index of the value.

        Returns
        -------
        The value at the index.
        """
        return getattr(
            self,
            self._make_key(index),
        )

    @classmethod
    def from_dict(
        cls,
        d,
        reference: Optional[Dict[str, str]] = None,
        loaded_ids: Optional[dict] = None,
    ) -> "Array":
        """
        Create an array from a dictionary.

        Parameters
        ----------
        d
            The dictionary.
        reference
            A dictionary of references.
        loaded_ids
            A dictionary of loaded ids.

        Returns
        -------
        The array.
        """
        arguments = d["arguments"]
        shape = from_dict(arguments["shape"])
        array = cls(shape)
        for key, value in arguments.items():
            if key.startswith("prior"):
                setattr(array, key, from_dict(value))

        return array

    def tree_flatten(self):
        """
        Flatten this array model as a PyTree.
        """
        members = [self[index] for index in self.indices]
        return members, (self.shape,)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten a PyTree into an array model.
        """
        (shape,) = aux_data
        instance = cls(shape)
        for index, child in zip(instance.indices, children):
            instance[index] = child

        return instance

    @property
    def prior_class_dict(self):
        return {
            **{
                prior: cls
                for prior_model in self.direct_prior_model_tuples
                for prior, cls in prior_model[1].prior_class_dict.items()
            },
            **{prior: np.ndarray for _, prior in self.direct_prior_tuples},
        }

    def gaussian_prior_model_for_arguments(self, arguments: Dict[Prior, Prior]):
        """
        Returns a new instance of model mapper with a set of Gaussian priors based on
        tuples provided by a previous nonlinear search.

        Parameters
        ----------
        arguments
            Tuples providing the mean and sigma of gaussians

        Returns
        -------
        A new model mapper populated with Gaussian priors
        """
        new_array = Array(self.shape)
        for index in self.indices:
            new_array[index] = self[index].gaussian_prior_model_for_arguments(arguments)

        return new_array



from typing import Union, Optional

from autofit.graphical.declarative.factor.hierarchical import HierarchicalFactor

from autofit.tools.namer import namer
from .abstract import AbstractDeclarativeFactor
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.samples.summary import SamplesSummary

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.prior_model import Model

from ...non_linear.combined_result import CombinedResult

class FactorGraphModel(AbstractDeclarativeFactor):
    def __init__(
        self,
        *model_factors: Union[AbstractDeclarativeFactor, HierarchicalFactor],
        name=None,
        include_prior_factors=True,
        use_jax : bool = False
    ):
        """
        A collection of factors that describe models, which can be
        used to create a graph and messages.

        If the models have shared priors then the graph has shared variables

        Parameters
        ----------
        model_factors
            Factors which are hierarchical or associated with a specific analysis
        """
        super().__init__(
            include_prior_factors=include_prior_factors,
            use_jax=use_jax,
        )
        self._model_factors = list(model_factors)
        self._name = name or namer(self.__class__.__name__)

    def tree_flatten(self):
        return (
            (self._model_factors,),
            (self._name, self._include_prior_factors),
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(
            *children[0],
            name=aux_data[0],
            include_prior_factors=aux_data[1],
        )

    @property
    def prior_model(self):
        """
        Construct a Collection comprising the prior models described
        in each model factor
        """
        from autofit.mapper.prior_model.collection import Collection

        return Collection(
            {factor.name: factor.prior_model for factor in self.model_factors}
        )

    @property
    def optimiser(self):
        raise NotImplemented()

    @property
    def info(self) -> str:
        """
        Info describing this collection.
        """
        return self.graph.info

    @property
    def name(self):
        return self._name

    def add(self, model_factor: AbstractDeclarativeFactor):
        """
        Add another factor to this collection.
        """
        self._model_factors.append(model_factor)

    def log_likelihood_function(self, instance: ModelInstance) -> float:
        """
        Compute the combined likelihood of each factor from a collection of instances
        with the same ordering as the factors.

        Parameters
        ----------
        instance
            A collection of instances, one corresponding to each factor

        Returns
        -------
        The combined likelihood of all factors
        """
        log_likelihood = 0
        for model_factor, instance_ in zip(self.model_factors, instance):
            log_likelihood += model_factor.log_likelihood_function(instance_)

        return log_likelihood

    @property
    def model_factors(self):
        model_factors = list()
        for model_factor in self._model_factors:
            if isinstance(model_factor, HierarchicalFactor):
                model_factors.extend(model_factor.factors)
            else:
                model_factors.append(model_factor)
        return model_factors

    def make_result(
        self,
        samples_summary: SamplesSummary,
        paths: AbstractPaths,
        samples: Optional[SamplesPDF] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[object] = None,
    ) -> CombinedResult:
        """
        Make a result from the samples summary and paths.

        The top level result accounts for the combined model.
        There is one child result for each model factor.

        Parameters
        ----------
        samples_summary
            A summary of the samples
        paths
            Handles saving and loading data
        samples
            The full list of samples
        search_internal
        analysis

        Returns
        -------
        A result with child results for each model factor
        """
        child_results = [
            model_factor.analysis.make_result(
                samples_summary=samples_summary.subsamples(
                    model_factor.prior_model,
                ),
                paths=paths,
                samples=samples.subsamples(model_factor.prior_model)
                if samples
                else None,
                search_internal=search_internal,
                analysis=model_factor,
            )
            for model_factor in self.model_factors
        ]
        return CombinedResult(
            child_results,
            samples_summary=samples_summary,
            paths=paths,
            samples=samples,
            search_internal=search_internal,
            analysis=analysis,
        )

    def _for_each_analysis(
        self,
        name,
        paths,
        *args,
        **kwargs,
    ):
        """
        Convenience function to call an underlying function for each
        analysis with a paths object with an integer attached to the
        end.

        Parameters
        ----------
        func
            Some function of the analysis class
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        """
        results = []
        for (i, analysis), *args in zip(
            enumerate(self.model_factors),
            *args,
        ):
            child_paths = paths.for_sub_analysis(analysis_name=f"analyses/analysis_{i}")
            func = getattr(analysis, name)
            results.append(
                func(
                    child_paths,
                    *args,
                    **kwargs,
                )
            )

        return results

    def visualize(
        self,
        paths: AbstractPaths,
        instance: ModelInstance,
        during_analysis: bool,
    ):
        """
        Visualise the instances provided using each factor.

        Instances in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        instance
            A collection of instances, each corresponding to a factor
        during_analysis
            Is this visualisation during analysis?
        """
        self._for_each_analysis(
            "visualize",
            paths,
            instance,
            during_analysis=during_analysis,
        )

    def visualize_before_fit(
        self,
        paths: AbstractPaths,
        model: Model,
    ):
        """
        Visualise the model provided using each factor.

        Models in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        model
            A collection of models, each corresponding to a factor
        """
        self._for_each_analysis(
            "visualize_before_fit",
            paths,
            model,
        )

    def save_attributes(self, paths: AbstractPaths):
        """
        Save the attributes of the analysis to the paths object.
        """
        self._for_each_analysis("save_attributes", paths)

    def save_results(self, paths: AbstractPaths, result):
        """
        Save the results of the analysis to the paths object.
        """
        self._for_each_analysis("save_results", paths, result)

    def visualize_combined(
        self,
        instance,
        paths: AbstractPaths,
        during_analysis,
    ):
        self.model_factors[0].visualize_combined(
            self.model_factors,
            paths,
            instance,
            during_analysis=during_analysis,
        )

    def perform_quick_update(self, paths, instance):

        try:
            self.model_factors[0].visualize_combined(
                analyses=self.model_factors,
                paths=paths,
                instance=instance,
                during_analysis=True,
                quick_update=True,
            )
        except Exception as e:
            pass


from typing import Optional, Dict

from autofit.non_linear.parallel import AbstractJob, AbstractJobResult


class JobResult(AbstractJobResult):
    def __init__(self, result, result_list_row, number):
        """
        The result of a job

        Parameters
        ----------
        result
            The result of a grid search
        result_list_row
            A row in the result list
        """
        super().__init__(number)
        self.result = result
        self.result_list_row = result_list_row


class Job(AbstractJob):
    def __init__(
        self,
        search_instance,
        model,
        analysis,
        arguments,
        index,
        info: Optional[Dict] = None,
    ):
        """
        A job to be performed in parallel.

        Parameters
        ----------
        search_instance
            An instance of an optimiser
        analysis
            An analysis
        arguments
            The grid search arguments
        """
        super().__init__(number=index)
        self.search_instance = search_instance
        self.analysis = analysis
        self.model = model
        self.arguments = arguments
        self.index = index
        self.info = info

    def perform(self):

        result = self.search_instance.fit(
            model=self.model,
            analysis=self.analysis,
            info=self.info,
        )
        result_list_row = [
            self.index,
            *[
                prior.lower_limit
                for prior in self.model.sort_priors_alphabetically(
                    self.arguments.values()
                )
            ],
            result.log_likelihood,
            result.samples_summary.log_evidence,
        ]

        return JobResult(result, result_list_row, self.number)



import os

from autoconf import conf

from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.database import DatabasePaths
from autofit.non_linear.paths.null import NullPaths

class Visualizer:

    def should_visualize(
        self, paths: AbstractPaths, during_analysis: bool = True
    ) -> bool:
        """
        Whether a visualize method should be called perform visualization, which depends on the following:

        1) If a model-fit has already completed, the default behaviour is for visualization to be bypassed in order
        to make model-fits run faster.

        2) If a model-fit has completed, but it is the final visualization output where `during_analysis` is False,
        it should be performed.

        3) Visualization can be forced to run via the `force_visualization_overwrite`, for example if a user
        wants to plot additional images that were not output on the original run.

        4) If the analysis is running a database session visualization is switched off.

        5) If PyAutoFit test mode is on visualization is disabled, irrespective of the `force_visualization_overwite`
        config input.

        Parameters
        ----------
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.

        Returns
        -------
        A bool determining whether visualization should be performed or not.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return False

        if isinstance(paths, DatabasePaths) or isinstance(paths, NullPaths):
            return False

        if conf.instance["general"]["output"]["force_visualize_overwrite"]:
            return True

        if not during_analysis:
            return True

        return not paths.is_complete

    """
    Methods associated with visualising analysis, model and data before, during
    or after an optimisation.
    """

    @staticmethod
    def visualize_before_fit(
        analysis,
        paths: AbstractPaths,
        model: AbstractPriorModel,
    ):
        pass

    @staticmethod
    def visualize(
        analysis,
        paths: AbstractPaths,
        instance,
        during_analysis,
    ):
        pass

    @staticmethod
    def visualize_before_fit_combined(
        analyses,
        paths: AbstractPaths,
        model: AbstractPriorModel,
    ):
        pass

    @staticmethod
    def visualize_combined(
        analyses,
        paths: AbstractPaths,
        instance,
        during_analysis,
    ):
        pass



import tempfile
from typing import Dict, Optional

from .abstract import AbstractPaths


class NullPaths(AbstractPaths):
    """
    Null version of paths object for avoiding writing of files to disk
    """

    @property
    def samples(self):
        return None

    def save_latent_samples(self, latent_samples):
        pass

    def save_json(self, name, object_dict: dict, prefix: str = ""):
        pass

    def load_json(self, name, prefix: str = "") -> dict:
        pass

    def save_array(self, name, array):
        pass

    def load_array(self, name):
        pass

    def save_fits(self, name: str, fits, prefix: str = ""):
        pass

    def load_fits(self, name: str, prefix: str = ""):
        pass

    def __init__(self):
        super().__init__()
        self.objects = dict()
        self._samples_path = tempfile.mkdtemp()

    def save_summary(
        self,
        samples,
        latent_samples,
        log_likelihood_function_time,
        visualization_time = None,
    ):
        pass

    @property
    def samples_path(self) -> str:
        return self._samples_path

    @AbstractPaths.parent.setter
    def parent(self, parent):
        pass

    @property
    def is_grid_search(self) -> bool:
        return False

    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "AbstractPaths":
        return NullPaths()

    def save_all(self, search_config_dict=None, info=None):
        pass

    def save_object(self, name: str, obj: object, prefix: str = ""):
        self.objects[name] = obj

    def load_object(self, name: str, prefix: str = ""):
        return self.objects[name]

    def remove_object(self, name: str):
        pass

    def is_object(self, name: str) -> bool:
        pass

    @property
    def is_complete(self) -> bool:
        return False

    def completed(self):
        pass

    def save_search_internal(self, obj):
        pass

    def load_search_internal(self):
        pass

    def remove_search_internal(self):
        pass

    @property
    def search_internal_path(self):
        pass

    def load_samples(self):
        pass

    def save_samples(self, samples):
        pass

    def load_samples_info(self):
        pass

    def zip_remove(self):
        pass



import copy
import itertools
from typing import Type, Union, Tuple, Optional, Dict
import logging

from autoconf.class_path import get_class
from autoconf.dictable import from_dict, to_dict
from .identifier import Identifier

logger = logging.getLogger(__name__)


def dereference(reference: Optional[dict], name: str):
    if reference is None:
        return None
    updated = {}
    for key, value in reference.items():
        array = key.split(".")
        if array[0] == name:
            updated[".".join(array[1:])] = value
    return updated


class ModelObject:
    _ids = itertools.count()

    @classmethod
    def next_id(cls):
        return next(cls._ids)

    def __init__(
        self,
        id_=None,
        label=None,
    ):
        """
        A generic object in AutoFit

        Parameters
        ----------
        id_
            A unique integer identifier. This is used to hash and order priors.
        label
            A label which can optionally be set for visualising this object in a
            graph.
        """
        self.id = int(self.next_id() if id_ is None else id_)
        self._label = label

    def replacing_for_path(self, path: Tuple[str, ...], value) -> "ModelObject":
        """
        Create a new model replacing the value for a given path with a new value

        Parameters
        ----------
        path
            A path indicating the sequence of names used to address an object
        value
            A value that should replace the object at the given path

        Returns
        -------
        A copy of this with an updated value
        """
        new = copy.deepcopy(self)
        obj = new
        for key in path[:-1]:
            if isinstance(key, int):
                obj = obj[key]
            else:
                obj = getattr(obj, key)

        key = path[-1]
        if isinstance(key, int):
            obj[key] = value
        else:
            setattr(obj, key, value)
        return new

    def has(self, cls: Union[Type, Tuple[Type, ...]]) -> bool:
        """
        Does this instance have an attribute which is of type cls?
        """
        for value in self.__dict__.values():
            if isinstance(value, cls):
                return True
        return False

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def component_number(self):
        return self.id

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    @property
    def identifier(self):
        return str(Identifier(self))

    @classmethod
    def from_dict(
        cls,
        d,
        reference: Optional[Dict[str, str]] = None,
        loaded_ids: Optional[dict] = None,
    ):
        """
        Recursively parse a dictionary returning the model, collection or
        instance that is represents.

        Parameters
        ----------
        d
            A dictionary representation of some object
        reference
            An optional dictionary mapping names to class paths. This is used
            to specify the type of a model or instance.

            Maps paths to class paths. For example:
            "path.in.model": "path.to.Class"

            In this case, the class path "path.to.Class" will be used to
            instantiate the object at "path.in.model". If no class path is
            specified, or no type can be found for the class path in 'd', then
            a Collection will be used as a placeholder.

            This is used to specify the type of a model or instance.
        loaded_ids
            A dictionary mapping ids to instances. This is used to ensure that
            all instances with the same id are the same object.

        Returns
        -------
        An instance
        """
        from autofit.mapper.prior_model.collection import Collection
        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.abstract import Prior
        from autofit.mapper.prior.tuple_prior import TuplePrior
        from autofit.mapper.prior.arithmetic.compound import Compound
        from autofit.mapper.prior.arithmetic.compound import ModifiedPrior
        from .prior.constant import Constant

        if isinstance(d, list):
            return [
                from_dict(
                    value,
                    reference=dereference(reference, str(index)),
                    loaded_ids=loaded_ids,
                )
                for index, value in enumerate(d)
            ]

        if not isinstance(d, dict):
            return d

        loaded_ids = {} if loaded_ids is None else loaded_ids

        type_ = d["type"]

        def get_class_path():
            try:
                return reference[""]
            except (KeyError, TypeError):
                return d.pop("class_path")

        if type_ == "model":
            class_path = get_class_path()
            try:
                instance = Model(get_class(class_path))
            except (ModuleNotFoundError, AttributeError):
                logger.warning(
                    f"Could not find type for class path {class_path}. Defaulting to Collection placeholder."
                )
                instance = Collection()
        elif type == "constant":
            return Constant(value=d["value"])
        elif type_ == "collection":
            instance = Collection()
        elif type_ == "tuple_prior":
            instance = TuplePrior()
        elif type_ == "compound":
            return Compound.from_dict(
                d,
                reference=dereference(reference, "assertion"),
                loaded_ids=loaded_ids,
            )
        elif type_ == "modified":
            return ModifiedPrior.from_dict(
                d,
                reference=dereference(reference, "prior"),
                loaded_ids=loaded_ids,
            )
        elif type_ == "dict":
            return {
                key: from_dict(
                    value,
                    reference=dereference(reference, key),
                    loaded_ids=loaded_ids,
                )
                for key, value in d["arguments"].items()
                if value
            }
        elif type_ == "instance":
            class_path = get_class_path()
            try:
                cls_ = get_class(class_path)
                # noinspection PyArgumentList
                return cls_(
                    **{
                        key: from_dict(
                            value,
                            reference=dereference(reference, key),
                            loaded_ids=loaded_ids,
                        )
                        for key, value in d["arguments"].items()
                    }
                )
            except (ModuleNotFoundError, AttributeError):
                from autofit.mapper.model import ModelInstance

                logger.warning(
                    f"Could not find type for class path {class_path}. Defaulting to Instance placeholder."
                )
                instance = ModelInstance()
        elif type_ == "array":
            from autofit.mapper.prior_model.array import Array

            return Array.from_dict(d)
        else:
            try:
                return Prior.from_dict(d, loaded_ids=loaded_ids)
            except KeyError:
                cls_ = get_class(type_)
                instance = object.__new__(cls_)

        for key, value in d["arguments"].items():
            try:
                setattr(
                    instance,
                    key,
                    from_dict(
                        value,
                        reference=dereference(reference, key),
                        loaded_ids=loaded_ids,
                    ),
                )
            except KeyError:
                pass

        if "assertions" in d:
            instance.assertions = [
                from_dict(
                    value,
                    reference=dereference(reference, "assertions"),
                    loaded_ids=loaded_ids,
                )
                for value in d["assertions"]
            ]

        return instance

    def dict(self) -> dict:
        """
        A dictionary representation of this object
        """
        from autofit.mapper.prior_model.abstract import AbstractPriorModel
        from autofit.mapper.prior_model.collection import Collection
        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.tuple_prior import TuplePrior
        from autofit.mapper.prior_model.array import Array
        from autofit.mapper.prior.constant import Constant

        if isinstance(self, Collection):
            type_ = "collection"
        elif isinstance(self, AbstractPriorModel) and self.prior_count == 0:
            type_ = "instance"
        elif isinstance(self, Model):
            type_ = "model"
        elif isinstance(self, TuplePrior):
            type_ = "tuple_prior"
        elif isinstance(self, Array):
            type_ = "array"
        elif isinstance(self, Constant):
            type_ = "constant"
        else:
            raise AssertionError(
                f"{self.__class__.__name__} cannot be serialised to dict"
            )

        try:
            assertions = [assertion.dict() for assertion in self._assertions]
        except AttributeError:
            assertions = []

        dict_ = {
            "type": type_,
        }

        if assertions:
            dict_["assertions"] = assertions

        arguments = {}

        for key, value in self._dict.items():
            try:
                value = to_dict(value)
            except AttributeError:
                pass
            except TypeError:
                pass
            arguments[key] = value

        dict_["arguments"] = arguments
        return dict_

    @property
    def _dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ("component_number", "item_number", "id", "cls", "label")
            and not key.startswith("_")
        }



from abc import abstractmethod
from typing import Dict, List, Optional
import warnings

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.samples.samples import Sample
from autofit.non_linear.samples.stored import SamplesStored

from autofit import exc


class SamplesNest(SamplesPDF):
    def __init__(
        self,
        model: AbstractPriorModel,
        sample_list: List[Sample],
        samples_info: Optional[Dict] = None,
    ):
        """
        Contains the samples of the non-linear search, including parameter values, log likelihoods,
        weights and other quantites.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        number_live_points
            The number of live points used by the nested sampler.
        samples_info
            Contains information on the samples (e.g. total iterations, time to run the search, etc.).
        """

        super().__init__(
            model=model,
            sample_list=sample_list,
            samples_info=samples_info,
        )

    def __add__(self, other: "SamplesNest") -> "SamplesNest":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Parameters
        ----------
        other
            Another Samples class

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        warnings.warn(
            f"Addition of {self.__class__.__name__} cannot retain results in native format. "
            "Visualization of summed samples diabled.",
            exc.SamplesWarning,
        )

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            samples_info=self.samples_info,
        )

    @property
    def number_live_points(self):
        return self.samples_info["number_live_points"]

    @property
    def log_evidence(self):
        return self.samples_info["log_evidence"]

    @property
    def total_samples(self):
        return self.samples_info["total_samples"]

    @property
    def total_accepted_samples(self) -> int:
        """
        The total number of accepted samples performed by the nested sampler.
        """
        return self.samples_info["total_accepted_samples"]

    @property
    def acceptance_ratio(self) -> float:
        """
        The ratio of accepted samples to total samples.
        """
        return self.total_accepted_samples / self.total_samples

    def samples_within_parameter_range(
        self, parameter_index: int, parameter_range: [float, float]
    ) -> "SamplesStored":
        """
        Returns a new set of Samples where all points without parameter values inside a specified range removed.

        For example, if our `Samples` object was for a model with 4 parameters are consistent of the following 3 sets
        of parameters:

        [[1.0, 2.0, 3.0, 4.0]]
        [[100.0, 2.0, 3.0, 4.0]]
        [[1.0, 2.0, 3.0, 4.0]]

        This function for `parameter_index=0` and `parameter_range=[0.0, 99.0]` would remove the second sample because
        the value 100.0 is outside the range 0.0 -> 99.0.

        Parameters
        ----------
        parameter_index
            The 1D index of the parameter (in the model's vector representation) whose values lie between the parameter
            range if a sample is kept.
        parameter_range
            The minimum and maximum values of the range of parameter values this parameter must lie between for it
            to be kept.
        """

        parameter_list = []
        log_likelihood_list = []
        log_prior_list = []
        weight_list = []

        for sample in self.sample_list:
            parameters = sample.parameter_lists_for_model(model=self.model)

            if (parameters[parameter_index] > parameter_range[0]) and (
                parameters[parameter_index] < parameter_range[1]
            ):
                parameter_list.append(parameters)
                log_likelihood_list.append(sample.log_likelihood)
                log_prior_list.append(sample.log_prior)
                weight_list.append(sample.weight)

        sample_list = Sample.from_lists(
            model=self.model,
            parameter_lists=parameter_list,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesStored(
            model=self.model,
            sample_list=sample_list,
            samples_info=self.samples_info,
        )



import csv
import logging
from typing import Tuple, Union

from pathlib import Path

from autoconf import conf
from autofit.tools.util import open_

logger = logging.getLogger(__name__)


class FormatNode:
    def __init__(self):
        self._dict = dict()
        self.value = None

    def __getitem__(self, item):
        if item not in self._dict:
            self._dict[item] = FormatNode()
        return self._dict[item]

    def __len__(self):
        return len(self._dict)

    def items(self):
        return self._dict.items()

    def list(self, indent=4, line_length=90):
        lines = []
        for key, value in self.items():
            indent_string = indent * " "
            if value.value is not None:
                value_string = str(value.value)
                space_string = max((line_length - len(str(key))), 1) * " "
                lines.append(f"{key}{space_string}{value_string}")

            if len(value) > 0:
                sub_lines = value.list(
                    indent=indent,
                    line_length=line_length - indent,
                )
                if value.value is None:
                    lines.append(key)
                for line in sub_lines:
                    lines.append(f"{indent_string}{line}")
        return lines


class TextFormatter:
    def __init__(self, line_length=90, indent=4):
        self.dict = FormatNode()
        self.line_length = line_length
        self.indent = indent

    def add_to_dict(self, path: Tuple[str, ...], value: str, info_dict: FormatNode):
        key = path[0]
        node = info_dict[key]
        if len(path) == 1:
            node.value = value
        else:
            self.add_to_dict(path[1:], value, node)

    def add(self, path: Tuple[str, ...], value):
        self.add_to_dict(path, value, self.dict)

    @property
    def text(self):
        return "\n".join(map(str, self.list))

    @property
    def list(self):
        return self.dict.list(
            indent=self.indent,
            line_length=self.line_length,
        )


def format_string_for_parameter_name(parameter_name: str) -> str:
    """
    Get the format for the label. Attempts to extract the key string associated with
    the dimension. Seems dodgy.

    Parameters
    ----------
    parameter_name
        A string label

    Returns
    -------
    format
        The format string (e.g {:.2f})
    """
    label_conf = conf.instance["notation"]["label_format"]

    try:
        # noinspection PyProtectedMember
        for key, value in sorted(
            label_conf["format"].items(),
            key=lambda item: len(item[0]),
            reverse=True,
        ):
            if key in parameter_name:
                return value
    except KeyError:
        pass

    logger.debug(
        "Could not find an entry for the parameter {} in the label_format.ini config at path {}".format(
            parameter_name, conf.instance.paths
        )
    )

    return "{:.4f}"


def convert_name_to_label(parameter_name, name_to_label):
    if not name_to_label:
        return parameter_name

    label_conf = conf.instance["notation"]["label"]

    try:
        return label_conf["label"][parameter_name]
    except KeyError:
        logger.debug(
            "Could not find an entry for the parameter {} in the label_format.iniconfig at paths {}".format(
                parameter_name, conf.instance.paths
            )
        )
        return parameter_name[0]


def add_whitespace(str0, str1, whitespace):
    return f"{str0}{str1.rjust(whitespace - len(str0) + len(str1))}"


def value_result_string_from(
    parameter_name, value, values_at_sigma=None, unit=None, format_string=None
):
    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if values_at_sigma is None:
        return f"{value}{unit}"
    else:
        lower_value_at_sigma = format_str.format(values_at_sigma[0])
        upper_value_at_sigma = format_str.format(values_at_sigma[1])
        return f"{value} ({lower_value_at_sigma}, {upper_value_at_sigma}){unit}"


def parameter_result_latex_from(
    parameter_name,
    value,
    errors=None,
    superscript="",
    unit=None,
    format_string=None,
    name_to_label=False,
    include_name=True,
    include_quickmath=False,
):
    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    name = convert_name_to_label(
        parameter_name=parameter_name, name_to_label=name_to_label
    )

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if not superscript:
        superscript = ""
    else:
        superscript = f"^{{\\rm{{{superscript}}}}}"

    if errors is None:
        if include_name:
            parameter_result_latex = f"{name}{superscript} = {value}{unit}"
        else:
            parameter_result_latex = f"{value}{unit}"

    else:
        lower_value_at_sigma = format_str.format(errors[0])
        upper_value_at_sigma = format_str.format(errors[1])

        parameter_result = (
            f"{value}^{{+{upper_value_at_sigma}}}_{{-{lower_value_at_sigma}}}{unit}"
        )

        if include_name:
            parameter_result_latex = f"{name}{superscript} = {value}^{{+{upper_value_at_sigma}}}_{{-{lower_value_at_sigma}}}{unit}"
        else:
            parameter_result_latex = parameter_result

    if "e" in format_str:
        psplit = parameter_result.split("e")

        parameter_result_latex = (
            f""
            f"{psplit[0]}"
            f"{psplit[1][3:]}"
            f"{psplit[2][3:]}"
            f"{psplit[3][-1]}"
            f" \\times 10^{{{int(psplit[1][1:3])}}}"
        )

    if not include_quickmath:
        return f"{parameter_result_latex} & "
    return f"${parameter_result_latex}$ & "


def output_list_of_strings_to_file(file, list_of_strings):
    with open_(file, "w") as f:
        f.write("".join(list_of_strings))


def write_table(headers, rows, filename: Union[str, Path]):
    """
    Write a table of parameters, posteriors, priors and likelihoods.

    Parameters
    ----------
    filename
        Where the table is to be written
    headers
        The headers of the table
    rows
        The rows of the table
    """
    column_max_widths = [
        max(len(str(value)) for value in column)
        for column in zip(*([headers] + list(rows)))
    ]

    with open(filename, "w+") as f:
        writer = csv.writer(f)

        def write_row(row_):
            writer.writerow(
                [
                    "{0:>{1}}".format("" if value is None else str(value), width)
                    for width, value in zip(column_max_widths, row_)
                ]
            )

        write_row(headers)
        for row in rows:
            write_row(row)



import shutil
from typing import Optional, Union

from autoconf.output import conditional_output
from autofit.database.sqlalchemy_ import sa
from .abstract import AbstractPaths
import numpy as np

from autofit.database.model import Fit
from autoconf.dictable import to_dict
from autofit.database.aggregator.info import Info


class DatabasePaths(AbstractPaths):
    def __init__(
        self,
        session,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths=True,
        parent=None,
        save_all_samples=False,
        unique_tag: Optional["str"] = None,
    ):
        super().__init__(
            name=name,
            path_prefix=path_prefix,
            is_identifier_in_paths=is_identifier_in_paths,
            parent=parent,
        )
        self.session = session
        self._fit = None
        self.save_all_samples = save_all_samples
        self.unique_tag = unique_tag

    __nullify_fields__ = ("session",)

    parent: "DatabasePaths"

    @AbstractPaths.parent.setter
    def parent(self, parent: "DatabasePaths"):
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.

        For database paths the parent must also be database paths.
        """
        if not (parent is None or isinstance(parent, DatabasePaths)):
            raise TypeError(
                "The parent of search that uses the database must also use the database"
            )
        self._parent = parent

    @property
    def is_grid_search(self) -> bool:
        return self.fit.is_grid_search

    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "DatabasePaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory. It also
        allows database fit objects to be related correctly.

        If no instance is set the prior median model is used
        to ensure that the parent object is queryable.

        Parameters
        ----------
        identifier
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.

        Returns
        -------
        A new paths object
        """
        self.fit.is_grid_search = True
        if self.fit.instance is None:
            self.fit.instance = self.model.instance_from_prior_medians()
        child = type(self)(
            session=self.session,
            name=name or self.name,
            path_prefix=path_prefix or self.path_prefix,
            is_identifier_in_paths=(
                is_identifier_in_paths
                if is_identifier_in_paths is not None
                else self.is_identifier_in_paths
            ),
            parent=self,
        )
        child.model = self.model
        child.search = self.search
        child._identifier = identifier
        return child

    def zip_remove(self):
        """
        Remove files from both the symlinked folder and the output directory
        """
        self.session.commit()
        Info(self.session).write()

        if self.remove_files:
            shutil.rmtree(self.output_path, ignore_errors=True)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d["session"]
        return d

    @conditional_output
    def save_json(self, name, object_dict: Union[dict, list], prefix: str = ""):
        """
        Save a dictionary as a json file in the database

        Parameters
        ----------
        name
            The name of the json
        object_dict
            The dictionary to save
        """
        self.fit.set_json(name, object_dict)

    def load_json(self, name: str, prefix: str = "") -> Union[dict, list]:
        """
        Load a json file from the database

        Parameters
        ----------
        name
            The name of the json

        Returns
        -------
        The loaded dictionary
        """
        return self.fit.get_json(name)

    @property
    def samples(self):
        return self.fit.samples

    @conditional_output
    def save_array(self, name, array: np.ndarray):
        """
        Save an array as a json file in the database

        Parameters
        ----------
        name
            The name of the array
        array
            The array to save
        """
        self.fit.set_array(name, array)

    def load_array(self, name: str) -> np.ndarray:
        """
        Load an array from the database

        Parameters
        ----------
        name
            The name of the array

        Returns
        -------
        The loaded array
        """
        return self.fit.get_array(name)

    @conditional_output
    def save_fits(self, name: str, fits, prefix: str = ""):
        """
        Save a fits file in the database

        Parameters
        ----------
        name
            The name of the fits file
        fits
            The fits file to save
        """
        self.fit.set_fits(name, fits)

    def load_fits(self, name: str, prefix: str = ""):
        """
        Load a fits file from the database

        Parameters
        ----------
        name
            The name of the fits file

        Returns
        -------
        The loaded hdu
        """
        return self.fit.get_fits(name)

    @conditional_output
    def save_object(self, name: str, obj: object, prefix: str = ""):
        self.fit[name] = obj

    def load_object(self, name: str, prefix: str = ""):
        return self.fit[name]

    def remove_object(self, name: str):
        del self.fit[name]

    def is_object(self, name: str) -> bool:
        return name in self.fit

    def save_search_internal(self, obj):
        pass

    def load_search_internal(self):
        pass

    def remove_search_internal(self):
        pass

    @property
    def fit(self) -> Fit:
        if self._fit is None:
            try:
                self._fit = (
                    self.session.query(Fit).filter(Fit.id == self.identifier).one()
                )
            except sa.orm.exc.NoResultFound:
                self._fit = Fit(
                    id=self.identifier,
                    is_complete=False,
                    unique_tag=self.unique_tag,
                    path_prefix=self.path_prefix,
                    name=self.name,
                )
                self.session.add(self._fit)

        if self.parent is not None:
            self._fit.parent = self.parent.fit
        return self._fit

    @property
    def is_complete(self) -> bool:
        return self.fit.is_complete

    def completed(self):
        self.fit.is_complete = True

    def save_summary(
        self,
        samples,
        latent_samples,
        log_likelihood_function_time,
        visualization_time = None,
    ):
        self.fit.instance = samples.max_log_likelihood()
        self.fit.max_log_likelihood = samples.max_log_likelihood_sample.log_likelihood

    def save_samples(self, samples):
        if not self.save_all_samples:
            samples = samples.minimise()

        self.fit.samples = samples
        self.fit.set_json("samples_info", samples.samples_info)

    def save_latent_samples(self, latent_samples):
        if not self.save_all_samples:
            latent_samples = latent_samples.minimise()
        self.fit.latent_samples = latent_samples

    def load_latent_samples(self):
        return self.fit.latent_samples

    def _load_samples(self):
        samples = self.fit.samples
        samples.model = self.model
        return samples

    def load_samples(self):
        return self._load_samples().sample_list

    def load_samples_info(self):
        return self._load_samples().samples_info

    def save_all(self, info, *_, **kwargs):
        self.fit.info = info
        self.fit.model = self.model
        if info:
            self.save_json("info", info)
        self.save_json("search", to_dict(self.search))
        self.save_json("model", to_dict(self.model))

        self.session.commit()
        Info(self.session).write()



#!/usr/bin/env python

"""
Filter and collate search outputs in all subdirectories.

Usage:

./aggregator.py (root_directory) [pipeline=pipeline phase=phase dataset=dataset]

Example:

./aggregator.py ../output pipeline=data_mass_x1_source_x1_positions
"""

import os
import zipfile
from collections import defaultdict
from os import path
from pathlib import Path
from shutil import rmtree
from typing import List, Union, Iterator, Optional

from .predicate import AttributePredicate
from .search_output import SearchOutput, GridSearchOutput, GridSearch


class AggregatorGroup:
    def __init__(self, groups: ["AggregatorGroup"]):
        """
        A group of aggregators produced by grouping search_outputs on a field.

        Parameters
        ----------
        groups
            Groups, each with a common value in the metadata file
        """
        self.groups = groups

    def filter(self, *predicates) -> "AggregatorGroup":
        """
        Apply predicates to the underlying groups whilst maintaining the total number of groups.

        Parameters
        ----------
        predicates
            Predicates that evaluate to `True` or `False` for any given phase.

        Returns
        -------
        A collection of groups of the same length with each group having the same or fewer members.
        """
        return AggregatorGroup([group.filter(*predicates) for group in self.groups])

    def __getitem__(self, item):
        return self.groups[item]

    def __len__(self):
        return len(self.groups)

    def values(self, name: str, parser=lambda o: o) -> List[List]:
        """
        Extract a list of lists values with a certain name from the output objects in
        this group.

        Parameters
        ----------
        name
            The name of the attribute to be extracted
        parser
            A function used to parse the output

        Returns
        -------
        A list of lists of values.
        """
        return [group.values(name, parser=parser) for group in self.groups]


def unzip_directory(directory: str):
    """
    Unzip all zip files in a directory recursively.
    """
    for root, _, filenames in os.walk(directory):
        for filename in filenames:
            if filename.endswith(".zip"):
                try:
                    with zipfile.ZipFile(path.join(root, filename), "r") as f:
                        f.extractall(path.join(root, filename[:-4]))
                except zipfile.BadZipFile:
                    raise zipfile.BadZipFile(
                        f"File is not a zip file: \n " f"{root} \n" f"{filename}"
                    )


def is_relative_to(path_a, path_b):
    """Return True if the path is relative to another path or False."""
    try:
        path_a.relative_to(path_b)
        return True
    except ValueError:
        return False


class Aggregator:
    def __init__(
        self,
        search_outputs: List[SearchOutput],
        grid_search_outputs: List[GridSearchOutput],
    ):
        """
        Class to aggregate phase results for all subdirectories in a given directory.

        Parameters
        ----------
        search_outputs
            A list of search_outputs
        """
        if len(search_outputs) > 20:
            print(
                "When aggregating many searches it can be more efficient to use the database.\n"
                "Checkout the database cookbook at this URL: "
                "https://pyautofit.readthedocs.io/en/latest/cookbooks/database.html"
            )
        self.search_outputs = search_outputs
        self.grid_search_outputs = grid_search_outputs

    def grid_searches(self):
        """
        A list of grid search outputs
        """
        return [
            GridSearch(
                output,
                [
                    search_output
                    for search_output in self.search_outputs
                    if is_relative_to(search_output.directory, output.directory)
                ],
            )
            for output in self.grid_search_outputs
        ]

    @classmethod
    def from_directory(
        cls,
        directory: Union[str, os.PathLike],
        completed_only=False,
        reference: Optional[dict] = None,
    ) -> "Aggregator":
        """
        Aggregate phase results for all subdirectories in a given directory.

        The whole directory structure is traversed and a Phase object created for each directory that contains a
        metadata file.

        Parameters
        ----------
        directory
            A directory in which the outputs of search_outputs are kept. This is searched recursively.
        completed_only
            If `True` only search_outputs with a .completed file (indicating the phase was completed)
            are included in the aggregator.
        reference
            A dictionary mapping paths to types to be used when loading models from disk.
        """
        print("Aggregator loading search_outputs... could take some time.")

        unzip_directory(directory)

        search_outputs = []
        grid_search_outputs = []

        for root, _, filenames in os.walk(directory):

            def should_add():
                return not completed_only or ".completed" in filenames

            if "metadata" in filenames:
                if should_add():
                    search_outputs.append(
                        SearchOutput(
                            Path(root),
                            reference=reference,
                        )
                    )
            if ".is_grid_search" in filenames:
                if should_add():
                    grid_search_outputs.append(
                        GridSearchOutput(
                            Path(root),
                        )
                    )

        if len(search_outputs) == 0:
            print(f"\nNo search_outputs found in {directory}\n")
        else:
            print(
                f"\n A total of {str(len(search_outputs))} search_outputs and results were found."
            )

        return cls(search_outputs, grid_search_outputs)

    def add_directory(self, directory: Union[str, Path]):
        """
        Add a directory to the aggregator.
        """
        aggregator = Aggregator.from_directory(directory)
        self.search_outputs.extend(aggregator.search_outputs)
        self.grid_search_outputs.extend(aggregator.grid_search_outputs)

    def remove_unzipped(self):
        """
        Removes the unzipped output directory for each phase.
        """
        for phase in self.search_outputs:
            split_path = path.split(phase.directory)[0]

            rmtree(split_path, ignore_errors=True)

    def __getitem__(self, item: Union[slice, int]) -> Union["Aggregator", SearchOutput]:
        """
        If an index is passed in then a specific phase output is returned.

        If a slice is passed in then an aggregator comprising several search_outputs is returned.

        Parameters
        ----------
        item
            A slice or index

        Returns
        -------
        An aggregator or phase
        """
        if isinstance(item, slice):
            return Aggregator(
                self.search_outputs[item],
                self.grid_search_outputs,
            )
        return self.search_outputs[item]

    def __len__(self):
        return len(self.search_outputs)

    def __iter__(self):
        return iter(self.search_outputs)

    def __getattr__(self, item):
        return AttributePredicate(item)

    def query(self, *predicates) -> "Aggregator":
        """
        Filter phase outputs by predicates. A predicate is created using a conditional
        operator.

        Another aggregator object is returned.

        Parameters
        ----------
        predicates
            Objects representing predicates that may evaluate to `True` or `False` for any
            given phase output.

        Returns
        -------
        An aggregator comprising all search_outputs that evaluate to `True` for all predicates.
        """
        search_outputs = self.search_outputs
        for predicate in predicates:
            search_outputs = predicate.filter(search_outputs)
        search_outputs = list(search_outputs)
        print(f"Filter found a total of {str(len(search_outputs))} results")
        return Aggregator(
            search_outputs=list(search_outputs),
            grid_search_outputs=self.grid_search_outputs,
        )

    def values(self, name: str, parser=lambda o: o) -> Iterator:
        """
        Get values from outputs with a given name.

        A list the same length as the number of search_outputs is returned
        where each item is the value of the attribute for a given
        phase.

        Parameters
        ----------
        name
            The name of an attribute expected to be associated with
            phase output. If a pickle file with this name is in the
            phase output directory then that pickle will be loaded.
        parser
            A function used to parse the output

        Returns
        -------
        A generator of values for the attribute
        """
        for value in map(lambda phase: getattr(phase, name), self.search_outputs):
            yield parser(value)

    def child_values(self, name: str) -> Iterator[List]:
        """
        Get values with a given name from the child analyses of each search in
        this aggregator.

        Parameters
        ----------
        name
            The name of an attribute expected to be associated with
            child analysis output. If a pickle file with this name
            is in the child analysis output directory then that pickle
            will be loaded.

        Returns
        -------
        A generator of values for the attribute
        """
        return (phase.child_values(name) for phase in self.search_outputs)

    def map(self, func):
        """
        Map some function onto the aggregated output objects.

        Parameters
        ----------
        func
            A function

        Returns
        -------
        A generator of results
        """
        return map(func, self.search_outputs)

    def group_by(self, field: str) -> AggregatorGroup:
        """
        Group the search_outputs by a field, e.g. pipeline.

        The object returned still permits filtering and attribute querying.

        Parameters
        ----------
        field
            The field by which to group

        Returns
        -------
        An object comprising lists of grouped fields
        """
        group_dict = defaultdict(list)
        for phase in self.search_outputs:
            group_dict[getattr(phase, field)].append(phase)
        return AggregatorGroup(list(map(Aggregator, group_dict.values())))

    @property
    def model_results(self) -> str:
        """
        A string joining headers and results for all included search_outputs.
        """
        return "\n\n".join(
            "{}\n\n{}".format(phase.header, phase.model_results)
            for phase in self.search_outputs
        )



from typing import Union, ItemsView, Any, Iterable, Tuple

from ..sqlalchemy_ import sa

from .model import Object


class Dict(Object):
    """
    Represents a Python dictionary in the database
    """

    __tablename__ = "dict"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "dict"}

    def __call__(self):
        d = {}
        for child in self.children:
            instance = child()
            if child.name != "":
                d[child.name] = instance
            else:
                d[instance[0]] = instance[1]

        return d

    @classmethod
    def _from_object(cls, source: dict):
        instance = cls()
        instance._add_children(source.items())
        instance.cls = dict
        return instance

    def _add_children(
        self, items: Union[ItemsView[str, Any], Iterable[Tuple[str, Any]]]
    ):
        """
        Add database representations of child attributes

        Parameters
        ----------
        items
            Attributes such as floats or priors that are associated
            with the real object
        """
        self.children = [
            Object.from_object(value, name=key)
            if isinstance(key, str)
            else Object.from_object((key, value))
            for key, value in items
        ]



from typing import Optional, Tuple

import numpy as np

from autofit.messages.normal import UniformNormalMessage
from autofit.messages.transform import log_10_transform, LinearShiftTransform
from .abstract import Prior
from ...messages.composed_transform import TransformedMessage

from autofit import exc

class LogUniformPrior(Prior):
    __identifier_fields__ = ("lower_limit", "upper_limit")
    __database_args__ = ("lower_limit", "upper_limit", "id_")

    def __init__(
        self,
        lower_limit: float = 1e-6,
        upper_limit: float = 1.0,
        id_: Optional[int] = None,
    ):
        """
        A prior with a log base 10 uniform distribution, defined between a lower limit and upper limit.

        The conversion of an input unit value, ``u``, to a physical value, ``p``, via the prior is as follows:

        .. math::

        For example for ``prior = LogUniformPrior(lower_limit=10.0, upper_limit=1000.0)``, an
        input ``prior.value_for(unit=0.5)`` is equal to 100.0.

        [Rich describe how this is done via message]

        Parameters
        ----------
        lower_limit
            The lower limit of the log10 uniform distribution defining the prior.
        upper_limit
            The upper limit of the log10 uniform distribution defining the prior.

        Examples
        --------

        prior = af.LogUniformPrior(lower_limit=0.0, upper_limit=2.0)

        physical_value = prior.value_for(unit=0.2)
        """

        self.lower_limit = float(lower_limit)
        self.upper_limit = float(upper_limit)

        if self.lower_limit <= 0.0:
            raise exc.PriorException(
                "The lower limit of a LogUniformPrior cannot be zero or negative."
            )
        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

        message = TransformedMessage(
            UniformNormalMessage,
            LinearShiftTransform(
                shift=np.log10(self.lower_limit),
                scale=np.log10(self.upper_limit / self.lower_limit),
            ),
            log_10_transform,
        )

        super().__init__(
            message=message,
            id_=id_,
        )

    def tree_flatten(self):
        return (
            self.lower_limit,
            self.upper_limit,
            self.id,
        ), ()

    @classmethod
    def with_limits(cls, lower_limit: float, upper_limit: float) -> "LogUniformPrior":
        """
        Create a new log 10 uniform prior centred between two limits
        with sigma distance between this limits.

        Note that these limits are not strict so exceptions will not
        be raised for values outside of the limits.

        This function is typically used in prior passing, where the
        result of a model-fit are used to create new Gaussian priors
        centred on the previously estimated median PDF model.

        Parameters
        ----------
        lower_limit
            The lower limit of the new LogUniform prior.
        upper_limit
            The upper limit of the new LogUniform Prior.

        Returns
        -------
        A new LogUniform.
        """
        return cls(
            lower_limit=max(0.000001, lower_limit),
            upper_limit=upper_limit,
        )

    __identifier_fields__ = ("lower_limit", "upper_limit")

    def log_prior_from_value(self, value, xp=np) -> float:
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
        posterior as log_prior + log_likelihood.

        This is used by certain non-linear searches (e.g. Emcee) in the log likelihood function evaluation.

        Parameters
        ----------
        value : float
            The physical value of this prior's corresponding parameter in a `NonLinearSearch` sample.
        """
        return 1.0 / value

    def value_for(self, unit: float) -> float:
        """
        Returns a physical value from an input unit value according to the limits of the log10 uniform prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        value
            The unit value mapped to a physical value according to the prior.

        Examples
        --------

        prior = af.LogUniformPrior(lower_limit=0.0, upper_limit=2.0)

        physical_value = prior.value_for(unit=0.2)
        """
        return super().value_for(unit)

    def dict(self) -> dict:
        """
        Return a dictionary representation of this GaussianPrior instance,
        including mean and sigma.

        Returns
        -------
        Dictionary containing prior parameters.
        """
        prior_dict = super().dict()
        return {**prior_dict, "lower_limit": self.lower_limit, "upper_limit": self.upper_limit}

    @property
    def limits(self) -> Tuple[float, float]:
        return self.lower_limit, self.upper_limit

    @property
    def parameter_string(self) -> str:
        return f"lower_limit = {self.lower_limit}, upper_limit = {self.upper_limit}"



from abc import ABC

import json
from copy import copy
import logging
import os
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from pathlib import Path

from autoconf import conf
from autoconf.class_path import get_class_path
from autofit import exc
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.sample import Sample

from .summary import SamplesSummary
from .interface import SamplesInterface, to_instance
from ...text.formatter import write_table

logger = logging.getLogger(__name__)


class Samples(SamplesInterface, ABC):
    def __init__(
        self,
        model: AbstractPriorModel,
        sample_list: List[Sample],
        samples_info: Optional[Dict] = None,
    ):
        """
        Contains the samples of the non-linear search, including parameter values, log likelihoods,
        weights and other quantites.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores samples of searches which provide maximum likelihood estimates of the  model-fit (e.g.
        PySwarms, LBFGS).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        samples_info
            Contains information on the samples (e.g. total iterations, time to run the search, etc.).
        """

        super().__init__(model=model)

        self.sample_list = sample_list
        self.samples_info = {
            **(samples_info or {}),
            "class_path": get_class_path(self.__class__),
        }

    def __str__(self):
        return f"{self.__class__.__name__}({len(self.sample_list)})"

    def __repr__(self):
        return str(self)

    @property
    def instances(self):
        """
        One model instance for each sample
        """
        return [
            self.model.instance_from_vector(
                sample.parameter_lists_for_paths(
                    self.paths if sample.is_path_kwargs else self.names
                ),
            )
            for sample in self.sample_list
        ]

    @property
    def log_evidence(self):
        return None

    @classmethod
    def from_list_info_and_model(
        cls,
        sample_list,
        samples_info,
        model: AbstractPriorModel,
    ):
        return cls(
            model=model,
            sample_list=sample_list,
            samples_info=samples_info,
        )

    def summary(self):
        return SamplesSummary(
            model=self.model,
            max_log_likelihood_sample=self.max_log_likelihood_sample,
        )

    def __add__(self, other: "Samples") -> "Samples":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
        )

    def __radd__(self, other):
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Overwriting `__radd__` enables the sum function to be used on a list of samples, e.g.:

        `samples = sum([samples_x5, samples_x5, samples_x5])`

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.

        Returns
        -------
        A class that combines the samples of a list of Samples objects.
        """
        return self

    def __len__(self):
        return len(self.sample_list)

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result._names = None
        result._paths = None
        return result

    def _check_addition(self, other: "Samples"):
        """
        When adding samples together, perform the following checks to make sure it is valid to add the two objects
        together:

        - That both objects being added are `Samples` objects.
        - That both models have the same prior count, else the dimensionality does not allow for valid addition.
        - That both `Samples` objects use an identical model, such that we are adding together the same parameters.

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.
        """

        def raise_exc():
            raise exc.SamplesException(
                "Cannot add together two Samples objects which have different models."
            )

        if not isinstance(self, Samples):
            raise_exc()

        if not isinstance(other, Samples):
            raise_exc()

        if self.model.prior_count != other.model.prior_count:
            raise_exc()

        for path_self, path_other in zip(self.model.paths, other.model.paths):
            if path_self != path_other:
                raise_exc()

    def values_for_path(self, path: Tuple[str]) -> List[float]:
        """
        Returns the value for a variable with a given path
        for each sample in the model
        """
        return [sample.kwargs[path] for sample in self.sample_list]

    @property
    def total_iterations(self) -> int:
        return self.samples_info["total_iterations"]

    @property
    def time(self) -> Optional[float]:
        return self.samples_info["time"]

    @property
    def parameter_lists(self):
        result = list()
        for sample in self.sample_list:
            tuples = self.paths if sample.is_path_kwargs else self.names
            result.append(sample.parameter_lists_for_paths(tuples))

        return result

    @property
    def total_samples(self):
        return len(self.sample_list)

    @property
    def weight_list(self):
        return [sample.weight for sample in self.sample_list]

    @property
    def log_likelihood_list(self):
        return [sample.log_likelihood for sample in self.sample_list]

    @property
    def log_posterior_list(self):
        return [sample.log_posterior for sample in self.sample_list]

    @property
    def log_prior_list(self):
        return [sample.log_prior for sample in self.sample_list]

    @property
    def parameters_extract(self):
        return np.asarray(self.parameter_lists).T

    @property
    def _headers(self) -> List[str]:
        """
        Headers for the samples table
        """

        return self.model.joined_paths + [
            "log_likelihood",
            "log_prior",
            "log_posterior",
            "weight",
        ]

    @property
    def _rows(self) -> List[List[float]]:
        """
        Rows in the samples table
        """

        log_likelihood_list = self.log_likelihood_list
        log_prior_list = self.log_prior_list
        log_posterior_list = self.log_posterior_list
        weight_list = self.weight_list

        for index, row in enumerate(self.parameter_lists):
            yield row + [
                log_likelihood_list[index],
                log_prior_list[index],
                log_posterior_list[index],
                weight_list[index],
            ]

    def write_table(self, filename: Union[str, Path]):
        """
        Write a table of parameters, posteriors, priors and likelihoods.

        Parameters
        ----------
        filename
            Where the table is to be written
        """
        write_table(
            filename=filename,
            headers=list(self._headers),
            rows=list(self._rows),
        )

    def info_to_json(self, filename):
        with open(filename, "w") as outfile:
            json.dump(self.samples_info, outfile)

    @property
    def max_log_likelihood_sample(self) -> Sample:
        """
        The index of the sample with the highest log likelihood.
        """
        most_likely_sample = None
        for sample in self.sample_list:
            if (
                most_likely_sample is None
                or sample.log_likelihood > most_likely_sample.log_likelihood
            ):
                most_likely_sample = sample
        return most_likely_sample

    @property
    def max_log_likelihood_index(self) -> int:
        """
        The index of the sample with the highest log likelihood.
        """
        return int(np.argmax(self.log_likelihood_list))

    @to_instance
    def max_log_likelihood(self) -> List[float]:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a model instance or
        list of values.
        """

        sample = self.max_log_likelihood_sample

        return sample.parameter_lists_for_paths(
            self.paths if sample.is_path_kwargs else self.names
        )

    @property
    def max_log_posterior_sample(self) -> Sample:
        return self.sample_list[self.max_log_posterior_index]

    @property
    def max_log_posterior_index(self) -> int:
        """
        The index of the sample with the highest log posterior.
        """
        return int(np.argmax(self.log_posterior_list))

    @to_instance
    def max_log_posterior(self) -> ModelInstance:
        """
        The parameters of the maximum log posterior sample of the `NonLinearSearch` returned as a model instance.
        """
        return self.parameter_lists[self.max_log_posterior_index]

    @to_instance
    def from_sample_index(self, sample_index: int) -> ModelInstance:
        """
        The parameters of an individual sample of the non-linear search, returned as a model instance.

        Parameters
        ----------
        sample_index
            The sample index of the weighted sample to return.
        """
        return self.parameter_lists[sample_index]

    def samples_above_weight_threshold_from(
        self, weight_threshold: Optional[float] = None, log_message: bool = False
    ) -> "Samples":
        """
        Returns a new `Samples` object containing only the samples with a weight above the input threshold.

        This function can be used after a non-linear search is complete, to reduce the samples to only the high weight
        values. The benefit of this is that the corresponding `samples.csv` file will be reduced in hard-disk size.

        For large libraries of results can significantly reduce the overall hard-disk space used and speed up the
        time taken to load the samples from a .csv file and perform analysis on them.

        For a sufficiently low threshold, this has a neglible impact on the numerical accuracy of the results, and
        even higher values can be used for aggresive use cases where hard-disk space is at a premium.

        Parameters
        ----------
        weight_threshold
            The threshold of weight at which a sample is included in the new `Samples` object.
        """

        if weight_threshold is None:
            weight_threshold = conf.instance["output"]["samples_weight_threshold"]

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            weight_threshold = None

        if weight_threshold is None:
            return self

        sample_list = []

        for sample in self.sample_list:
            if sample.weight > weight_threshold:
                sample_list.append(sample)

        if log_message:
            logger.info(
                f"Samples with weight less than {weight_threshold} removed from samples.csv."
            )

        return self.__class__(
            model=self.model,
            sample_list=sample_list,
            samples_info=self.samples_info,
        )

    def minimise(self) -> "Samples":
        """
        A copy of this object with only important samples retained
        """
        samples = copy(self)
        samples.sample_list = list(
            {self.max_log_likelihood_sample, self.max_log_posterior_sample}
        )
        return samples

    def with_paths(self, paths: Union[List[Tuple[str, ...]], List[str]]) -> "Samples":
        """
        Create a copy of this object with only attributes specified
        by a list of paths.

        Parameters
        ----------
        paths
            A list of paths to attributes. Only kwargs and model components
            specified by these paths are retained.

            All children of a given path are retained.

        Returns
        -------
        A set of samples with a reduced set of attributes
        """
        with_paths = copy(self)
        with_paths.model = self.model.with_paths(paths)
        with_paths.sample_list = [
            sample.with_paths(paths) for sample in self.sample_list
        ]
        return with_paths

    def without_paths(
        self, paths: Union[List[Tuple[str, ...]], List[str]]
    ) -> "Samples":
        """
        Create a copy of this object with only attributes not specified
        by a list of paths.

        Parameters
        ----------
        paths
            A list of paths to attributes. kwargs and model components
            specified by these paths are removed.

            All children of a given path are removed.

        Returns
        -------
        A set of samples with a reduced set of attributes
        """
        with_paths = copy(self)
        with_paths.model = self.model.without_paths(paths)
        with_paths.sample_list = [
            sample.without_paths(paths) for sample in self.sample_list
        ]
        return with_paths

    def subsamples(self, model):
        if self.model is None:
            return None

        path_map = self.path_map_for_model(model)
        copied = copy(self)
        copied._paths = None
        copied._names = None
        copied.model = model

        copied.sample_list = [sample.subsample(path_map) for sample in self.sample_list]
        return copied



from numbers import Number

from autofit.mapper.model_object import ModelObject


class Constant(float, ModelObject):
    def __new__(cls, value: float, id_=None):
        obj = super().__new__(cls, value)
        return obj

    def __init__(self, value: float, id_=None):
        """
        Represents a constant value in a model.

        This is equivalent to a prior.

        Note that two different instances of this class with the same value are not equal,
        but are equal once instantiated or in comparison to a float or int.

        Parameters
        ----------
        value
            The constant value.
        """
        if isinstance(value, Constant):
            value = value.value

        ModelObject.__init__(self, id_=id_)
        self.value = value

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, ModelObject):
            return ModelObject.__eq__(self, other)
        return self.value == other

    def __ne__(self, other):
        if isinstance(other, ModelObject):
            return ModelObject.__ne__(self, other)
        return self.value != other

    def __hash__(self):
        return hash(self.id)

    def dict(self):
        return {"type": "Constant", "value": self.value}



from abc import ABC
from pathlib import Path
from typing import cast

from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.paths.database import DatabasePaths
from autofit.non_linear.paths.directory import DirectoryPaths


class SubDirectoryPaths(ABC):
    analysis_name: str
    parent: AbstractPaths

    def __new__(
            cls,
            parent=None,
            analysis_name=None,
            is_flat=False,
    ):
        """
        Determine which kind of sub-directory paths to generate
        """
        if isinstance(
                parent,
                DatabasePaths
        ):
            return super(SubDirectoryPaths, cls).__new__(SubDirectoryPathsDatabase)
        return super(SubDirectoryPaths, cls).__new__(SubDirectoryPathsDirectory)

    @property
    def _output_path(self) -> Path:
        """
        The output path is customised to place output in a named directory in
        the analyses directory.
        """
        return self.parent.output_path / self.analysis_name


class SubDirectoryPathsDirectory(
    DirectoryPaths,
    SubDirectoryPaths
):
    def __init__(
            self,
            parent: DirectoryPaths,
            analysis_name: str,
            is_flat=False,
    ):
        """
        Paths for a child directory for a search which generates multiple searches
        such as in sensitivity mapping or graphical modelling

        Parameters
        ----------
        parent
            A paths object for the original search
        analysis_name
            The name of the analysis (used to generate a directory name)
        is_flat
            If true then any children generated retain the top level parent
            rather than recursively taking a sub-directory parent.
        """
        self.analysis_name = analysis_name
        super().__init__()
        if is_flat and isinstance(
                parent,
                SubDirectoryPaths
        ):
            self.parent = parent.parent
        else:
            self.parent = parent

    @property
    def output_path(self) -> Path:
        """
        The output path is customised to place output in a named directory in
        the analyses directory.
        """
        return self._output_path


class SubDirectoryPathsDatabase(
    DatabasePaths,
    SubDirectoryPaths
):
    def __init__(
            self,
            parent: DatabasePaths,
            analysis_name: str,
            is_flat=False,
    ):
        """
        Paths for a child directory for a search which generates multiple searches
        such as in sensitivity mapping or graphical modelling. Uses database to save
        most output data.

        Parameters
        ----------
        parent
            A paths object for the original search
        analysis_name
            The name of the analysis (used to generate a directory name)
        is_flat
            If true then any children generated retain the top level parent
            rather than recursively taking a sub-directory parent.
        """
        self.analysis_name = analysis_name
        super().__init__(
            parent.session
        )
        if is_flat and isinstance(
                parent,
                SubDirectoryPaths
        ):
            self.parent = cast(
                DatabasePaths,
                parent.parent
            )
        else:
            self.parent = parent

    @property
    def output_path(self) -> Path:
        """
        The output path is customised to place output in a named directory in
        the analyses directory.
        """
        return self._output_path



from typing import Optional, Tuple

from autofit.messages.truncated_normal import TruncatedNormalMessage
from .abstract import Prior


class TruncatedGaussianPrior(Prior):
    __identifier_fields__ = ("mean", "sigma", "lower_limit", "upper_limit")
    __database_args__ = ("mean", "sigma", "lower_limit", "upper_limit", "id_")

    def __init__(
        self,
        mean: float,
        sigma: float,
        lower_limit: float = float("-inf"),
        upper_limit: float = float("inf"),
        id_: Optional[int] = None,
    ):
        """
        A Gaussian prior defined by a normal distribution with optional truncation limits.

        This prior represents a Gaussian (normal) distribution with mean `mean`
        and standard deviation `sigma`, optionally truncated between `lower_limit`
        and `upper_limit`. The transformation from a unit interval input `u ∈ [0, 1]`
        to a physical parameter value `p` uses the inverse error function (erfcinv) applied
        to the Gaussian CDF, adjusted for truncation:

        .. math::
            p = \mu + \sigma \sqrt{2} \, \mathrm{erfcinv}(2 \times (1 - u))

        where :math:`\mu` is the mean and :math:`\sigma` the standard deviation.

        If truncation limits are specified, values outside the interval
        [`lower_limit`, `upper_limit`] are disallowed and the distribution is
        normalized over this interval.

        Parameters
        ----------
        mean
            The mean (center) of the Gaussian prior distribution.
        sigma
            The standard deviation (spread) of the Gaussian prior.
        lower_limit : float, optional
            The lower truncation limit (default: -∞).
        upper_limit : float, optional
            The upper truncation limit (default: +∞).
        id_ : Optional[int], optional
            Optional identifier for the prior instance.

        Examples
        --------
        Create a TruncatedGaussianPrior with mean 1.0, sigma 2.0, truncated between 0.0 and 2.0:

        >>> prior = TruncatedGaussianPrior(mean=1.0, sigma=2.0, lower_limit=0.0, upper_limit=2.0)
        >>> physical_value = prior.value_for(unit=0.5)  # Returns a value near 1.0 (mean)
        """
        super().__init__(
            message=TruncatedNormalMessage(
                mean=mean,
                sigma=sigma,
                lower_limit=float(lower_limit),
                upper_limit=float(upper_limit),
            ),
            id_=id_,
        )

    def tree_flatten(self):
        return (self.mean, self.sigma, self.lower_limit, self.upper_limit, self.id), ()

    @classmethod
    def with_limits(cls, lower_limit: float, upper_limit: float) -> "TruncatedGaussianPrior":
        """
        Create a new truncated gaussian prior centred between two limits
        with sigma distance between this limits.

        Note that these limits are not strict so exceptions will not
        be raised for values outside of the limits.

        This function is typically used in prior passing, where the
        result of a model-fit are used to create new Gaussian priors
        centred on the previously estimated median PDF model.

        Parameters
        ----------
        lower_limit
            The lower limit of the new Gaussian prior.
        upper_limit
            The upper limit of the new Gaussian Prior.

        Returns
        -------
        A new prior instance centered between the limits.
        """
        return cls(
            mean=(lower_limit + upper_limit) / 2,
            sigma=(upper_limit - lower_limit),
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    def dict(self) -> dict:
        """
        Return a dictionary representation of this GaussianPrior instance,
        including mean and sigma.

        Returns
        -------
        Dictionary containing prior parameters.
        """
        prior_dict = super().dict()
        return {
            **prior_dict,
            "mean": self.mean,
            "sigma": self.sigma,
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit
        }

    @property
    def limits(self) -> Tuple[float, float]:
        return self.lower_limit, self.upper_limit

    @property
    def parameter_string(self) -> str:
        """
        Return a human-readable string summarizing the GaussianPrior parameters.
        """
        return (f"mean = {self.mean}, "
                f"sigma = {self.sigma}, "
                f"lower_limit = {self.lower_limit}, "
                f"upper_limit = {self.upper_limit}"
                )



from typing import List, Optional, Union, Iterable, Tuple

import numpy as np

from autofit import exc
from autofit.non_linear.search.abstract_search import NonLinearSearch
from autofit.non_linear.grid.grid_list import GridList, as_grid_list
from autofit.mapper import model_mapper as mm
from autofit.mapper.prior.abstract import Prior

from autofit.non_linear.samples.interface import SamplesInterface


class AbstractGridSearchResult:
    def __init__(self, samples: GridList):
        self.samples = samples

    # noinspection PyTypeChecker
    @as_grid_list
    def physical_centres_lists_from(
        self,
        path: Union[str, Tuple[str, ...]],
    ) -> GridList:
        """
        Get the physical centres of the grid search from a path to an attribute of the instance in the samples.

        Parameters
        ----------
        path
            The path to the attribute to get from the instance

        Returns
        -------
        A list of lists of physical values
        """
        return self._physical_centres_lists_from(self.samples, path)

    # noinspection PyTypeChecker
    @as_grid_list
    def _physical_centres_lists_from(
        self,
        samples: GridList,
        path: Union[str, Tuple[str, ...]],
    ) -> GridList:
        """
        Get the physical centres of the grid search from a path to an attribute of the instance in the samples.

        Parameters
        ----------
        path
            The path to the attribute to get from the instance

        Returns
        -------
        A list of lists of physical values
        """
        if isinstance(path, str):
            path = path.split(".")

            def value_for_samples(samples):
                return samples.model.object_for_path(path).mean

        else:
            paths = [p.split(".") for p in path]

            def value_for_samples(samples):
                return tuple(samples.model.object_for_path(p).mean for p in paths)

        return [value_for_samples(samples) for samples in samples]


# noinspection PyTypeChecker
class GridSearchResult(AbstractGridSearchResult):
    def __init__(
        self,
        samples: List[SamplesInterface],
        lower_limits_lists: Union[List, GridList],
        grid_priors: List[Prior],
        parent: Optional[NonLinearSearch] = None,
    ):
        """
        The sample of a grid search.

        Parameters
        ----------
        samples
            The samples of the non linear optimizations performed at each grid step
        lower_limits_lists
            A list of lists of values representing the lower bounds of the grid searched values at each step
        """
        self.no_dimensions = len(lower_limits_lists[0])
        self.no_steps = len(lower_limits_lists)

        self.lower_limits_lists = GridList(lower_limits_lists, self.shape)
        self.side_length = int(self.no_steps ** (1 / self.no_dimensions))
        self.step_size = 1 / self.side_length
        self.grid_priors = grid_priors

        self.parent = parent

        super().__init__(GridList(samples, self.shape) if samples is not None else None)

    @property
    @as_grid_list
    def physical_lower_limits_lists(self) -> GridList:
        """
        The lower physical values for each grid square
        """
        return self._physical_values_for(self.lower_limits_lists)

    @property
    @as_grid_list
    def physical_centres_lists(self) -> GridList:
        """
        The middle physical values for each grid square
        """
        return self._physical_values_for(self.centres_lists)

    @property
    @as_grid_list
    def physical_upper_limits_lists(self) -> GridList:
        """
        The upper physical values for each grid square
        """
        return self._physical_values_for(self.upper_limits_lists)

    @property
    @as_grid_list
    def upper_limits_lists(self) -> GridList:
        """
        The upper values for each grid square
        """
        return [
            [limit + self.step_size for limit in limits]
            for limits in self.lower_limits_lists
        ]

    @property
    @as_grid_list
    def centres_lists(self) -> List:
        """
        The centre values for each grid square
        """
        return [
            [(upper + lower) / 2 for upper, lower in zip(upper_limits, lower_limits)]
            for upper_limits, lower_limits in zip(
                self.lower_limits_lists, self.upper_limits_lists
            )
        ]

    def _physical_values_for(self, unit_lists: GridList) -> List:
        """
        Compute physical values for lists of lists of unit hypercube
        values.

        Parameters
        ----------
        unit_lists
            A list of lists of hypercube values

        Returns
        -------
        A list of lists of physical values
        """
        return [
            [prior.value_for(limit) for prior, limit in zip(self.grid_priors, limits)]
            for limits in unit_lists
        ]

    def __setstate__(self, state):
        return self.__dict__.update(state)

    def __getstate__(self):
        return self.__dict__

    def __getattr__(self, item: str) -> object:
        """
        We default to getting attributes from the best sample. This allows promises to reference best samples.
        """
        return getattr(self.best_samples, item)

    @property
    def shape(self):
        return self.no_dimensions * (int(self.no_steps ** (1 / self.no_dimensions)),)

    @property
    def best_samples(self):
        """
        The best sample of the grid search. That is, the sample output by the non linear search that had the highest
        maximum figure of merit.

        Returns
        -------
        best_sample: sample
        """
        return max(
            self.samples,
            key=lambda sample: sample.log_likelihood,
        )

    @property
    def best_model(self):
        """
        Returns
        -------
        best_model: mm.ModelMapper
            The model mapper instance associated with the highest figure of merit from the grid search
        """
        return self.best_sample.model

    @property
    def all_models(self):
        """
        Returns
        -------
        all_models: [mm.ModelMapper]
            All model mapper instances used in the grid search
        """
        return [sample.model for sample in self.samples]

    @property
    def physical_step_sizes(self):
        physical_step_sizes = []

        # TODO : Make this work for all dimensions in a less ugly way.

        for dim in range(self.no_dimensions):
            values = [value[dim] for value in self.physical_lower_limits_lists]
            diff = [abs(values[n] - values[n - 1]) for n in range(1, len(values))]

            if dim == 0:
                physical_step_sizes.append(np.max(diff))
            elif dim == 1:
                physical_step_sizes.append(np.min(diff))
            else:
                raise exc.GridSearchException(
                    "This feature does not support > 2 dimensions"
                )

        return tuple(physical_step_sizes)

    @as_grid_list
    def attribute_grid(self, attribute_path: Union[str, Iterable[str]]) -> GridList:
        """
        Get a list of the attribute of the best instance from every search in a numpy array with the native dimensions
        of the grid search.

        Parameters
        ----------
        attribute_path
            The path to the attribute to get from the instance

        Returns
        -------
        A numpy array of the attribute of the best instance from every search in the grid search.
        """
        if isinstance(attribute_path, str):
            attribute_path = attribute_path.split(".")

        attribute_list = []
        for sample in self.samples:
            attribute = sample.instance
            for attribute_name in attribute_path:
                attribute = getattr(attribute, attribute_name)
            attribute_list.append(attribute)

        return attribute_list

    @as_grid_list
    def log_likelihoods(
        self,
        relative_to_value: float = 0.0,
    ) -> GridList:
        """
        The maximum log likelihood of every grid search on a NumPy array whose shape is the native dimensions of the
        grid search.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's maximum likelihood (corresponding to unit priors (0.0, 0.0)) are in the first
        value (E.g. entry [0, 0]) of the NumPy array.

        Parameters
        ----------
        relative_to_value
            The value to subtract from every log likelihood, for example if Bayesian model comparison is performed
            on the grid search and the subtracted value is the maximum log likelihood of a previous search.
        """
        return [sample.log_likelihood - relative_to_value for sample in self.samples]

    @as_grid_list
    def log_evidences(self, relative_to_value: float = 0.0) -> GridList:
        """
        The maximum log evidence of every grid search on a NumPy array whose shape is the native dimensions of the
        grid search.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's maximum evidence (corresponding to unit priors (0.0, 0.0)) are in the first
        value (E.g. entry [0, 0]) of the NumPy array.

        Parameters
        ----------
        relative_to_value
            The value to subtract from every log likelihood, for example if Bayesian model comparison is performed
            on the grid search and the subtracted value is the maximum log likelihood of a previous search.
        """
        return [sample.log_evidence - relative_to_value for sample in self.samples]

    def figure_of_merits(
        self, use_log_evidences: bool, relative_to_value: float = 0.0
    ) -> GridList:
        """
        Convenience method to get either the log likelihoods or log evidences of the grid search.

        Parameters
        ----------
        use_log_evidences
            If true, the log evidences are returned, otherwise the log likelihoods are returned.
        relative_to_value
            The value to subtract from every log likelihood, for example if Bayesian model comparison is performed
            on the grid search and the subtracted value is the maximum log likelihood of a previous search.
        """

        if use_log_evidences:
            return self.log_evidences(relative_to_value)
        return self.log_likelihoods(relative_to_value)



import numpy as np
from functools import wraps
import logging
import os

from autoconf import conf

from autofit.non_linear.plot.output import Output

logger = logging.getLogger(__name__)

def skip_plot_in_test_mode(func):
    """
    Skips visualization plots of non-linear searches if test mode is on.

    Parameters
    ----------
    func
        A function which plots a result of a non-linear search.

    Returns
    -------
        A function that plots a visual, or None if test mode is on.
    """

    @wraps(func)
    def wrapper(
        *args,
        **kwargs
    ):
        """
        Skips visualization plots of non-linear searches if test mode is on.

        Parameters
        ----------
        obj
            An plotter object which performs visualization of a non-linear search.

        Returns
        -------
            A function that plots a visual, or None if test mode is on.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return

        return func(*args, **kwargs)

    return wrapper



class SamplesPlotter:
    def __init__(
            self, 
            samples,
            output : Output = Output()
    ):

        self.samples = samples
        self.output = output

    @property
    def model(self):
        return self.samples.model

    @property
    def log_posterior_list(self):
        return self.samples.log_posterior_list

    def close(self):

        import matplotlib.pyplot as plt

        if plt.fignum_exists(num=1):
            plt.clf()
            plt.close()

    def log_plot_exception(self, plot_name : str):
        """
        Plotting the results of a ``dynesty`` model-fit before they have converged on an
        accurate estimate of the posterior can lead the ``dynesty`` plotting routines
        to raise a ``ValueError``.

        This exception is caught in each of the plotting methods below, and this
        function is used to log the behaviour.

        Parameters
        ----------
        plot_name
            The name of the ``dynesty`` plot which raised a ``ValueError``
        """

        logger.info(
            f"""{self.__class__.__name__} unable to produce {plot_name} visual: posterior estimate therefore
            not yet sufficient for this model-fit is not yet robust enough to do this. 
            Visuals should be produced in later update, once posterior estimate is updated.
            """
        )

    def corner_cornerpy(self, **kwargs):
        """
        Plots a corner plot via the visualization library `corner.py`.

        This plots a corner plot including the 1-D and 2-D marginalized posteriors.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return

        import matplotlib.pylab as pylab

        config_dict = conf.instance["visualize"]["plots_settings"]["corner_cornerpy"]

        params = {'font.size' : int(config_dict["fontsize"])}
        pylab.rcParams.update(params)

        import corner

        corner.corner(
            data=np.asarray(self.samples.parameter_lists),
            weight_list=self.samples.weight_list,
            labels=self.model.parameter_labels_with_superscripts_latex,
        )

        self.output.to_figure(auto_filename="corner")
        self.close()






import datetime as dt
import os
import time


class Timer:

    def __init__(self, timer_path: str):
        """Times the run-time of the non-linear searches, by outputting a start-time file to the hard-disk and using
        this to determine the total run time when a `NonLinearSearch` update is performed.

        Parameters
        ----------
        timer_path
            The directory in which the timer should save results
        """

        self.timer_path = timer_path
        os.makedirs(
            timer_path,
            exist_ok=True
        )

    def start(self):
        """
        Record the start time of a `NonLinearSearch` as universal date time, so that the run-time of the search can be
        recorded.
        """

        start_time_path = self.timer_path / ".start_time"

        try:
            with open(start_time_path) as f:
                float(f.read())
        except FileNotFoundError:
            start = time.time()
            with open(start_time_path, "w+") as f:
                f.write(str(start))

    def update(self):
        """
        Update the timer of the `NonLinearSearch` so it reflections how long the `NonLinearSearch` ahs been running.
        """

        try:
            execution_time = str(time.time() - float(self.start_time))
        except TypeError:
            return

        with open(
                self.timer_path / ".time", "w+"
        ) as f:
            f.write(execution_time)

    @property
    def start_time(self):
        """
        Load the start time written to hard disk from the .start_time file.
        """
        try:
            with open(
                    self.timer_path / ".start_time", "r"
            ) as f:
                return f.read()
        except FileNotFoundError:
            return None

    @property
    def time(self):
        """
        Load the total time of the `NonLinearSearch` written to hard disk fom the .start_time file.
        """
        try:
            with open(
                    self.timer_path / ".time", "r"
            ) as f:
                return f.read()
        except FileNotFoundError:
            return None



import os
import numpy as np
from typing import List

import autofit as af


class VisualizerExample(af.Visualizer):
    """
    Methods associated with visualising analysis, model and data before, during
    or after an optimisation.
    """

    @staticmethod
    def visualize_before_fit(
        analysis,
        paths: af.AbstractPaths,
        model: af.AbstractPriorModel,
    ):
        """
        Before a model-fit begins, the `visualize_before_fit` method is called and is used to output images
        of quantities that do not change during the fit (e.g. the data).

        The function receives as input an instance of the `Analysis` class which is being used to perform the fit,
        which is used to perform the visualization (e.g. it contains the data and noise map which are plotted).

        For your model-fitting problem this function will be overwritten with plotting functions specific to your
        problem.

        Parameters
        ----------
        analysis
            The analysis class used to perform the model-fit whose quantities are being visualized.
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        model
            The model which is fitted to the data, which may be used to customize the visualization.
        """

        import matplotlib.pyplot as plt

        xvalues = np.arange(analysis.data.shape[0])

        plt.errorbar(
            x=xvalues,
            y=analysis.data,
            yerr=analysis.noise_map,
            color="k",
            ecolor="k",
            elinewidth=1,
            capsize=2,
        )
        plt.title("The 1D Dataset.")
        plt.xlabel("x values of profile")
        plt.ylabel("Profile normalization")

        os.makedirs(paths.image_path, exist_ok=True)
        plt.savefig(paths.image_path / "data.png")
        plt.clf()
        plt.close()

    @staticmethod
    def visualize(
        analysis,
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis : bool
    ):
        """
        During a model-fit, the `visualize` method is called throughout the non-linear search and is used to output
        images indicating the quality of the fit so far.

        The function receives as input an instance of the `Analysis` class which is being used to perform the fit,
        which is used to perform the visualization (e.g. it generates the model data which is plotted).

        The `instance` passed into the visualize method is maximum log likelihood solution obtained by the model-fit
        so far which can output on-the-fly images showing the best-fit model so far.

        For your model-fitting problem this function will be overwritten with plotting functions specific to your
        problem.

        Parameters
        ----------
        analysis
            The analysis class used to perform the model-fit whose quantities are being visualized.
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        instance
            An instance of the model that is being fitted to the data by this analysis (whose parameters have been set
            via a non-linear search).
        during_analysis
            If True the visualization is being performed midway through the non-linear search before it is finished,
            which may change which images are output.
        """

        import matplotlib.pyplot as plt

        xvalues = np.arange(analysis.data.shape[0])
        model_data_1d_list = []

        try:
            for profile in instance:
                try:
                    model_data_1d_list.append(profile.model_data_from(xvalues=xvalues))
                except AttributeError:
                    pass
        except TypeError:
            model_data_1d_list.append(instance.model_data_from(xvalues=xvalues))

        model_data_1d = sum(model_data_1d_list)

        plt.errorbar(
            x=xvalues,
            y=analysis.data,
            yerr=analysis.noise_map,
            color="k",
            ecolor="k",
            elinewidth=1,
            capsize=2,
        )
        plt.plot(xvalues, model_data_1d, color="r")

        for model_data_1d_profile in model_data_1d_list:

            plt.plot(xvalues, model_data_1d_profile, color="b", linestyle="--")

        plt.title("Model fit to multiple 1D profiles dataset.")
        plt.xlabel("x values of profile")
        plt.ylabel("Profile normalization")

        os.makedirs(paths.image_path, exist_ok=True)
        plt.savefig(paths.image_path / "model_fit.png")
        plt.clf()
        plt.close()

    @staticmethod
    def visualize_before_fit_combined(
        analyses,
        paths: af.AbstractPaths,
        model: af.AbstractPriorModel,
    ):
        """
        Multiple instances of the `Analysis` class can be summed together, meaning that the model is fitted to all
        datasets simultaneously via a summed likelihood function.

        The function receives as input a list of instances of every `Analysis` class which is being used to perform
        the summed analysis fit. This is used which is used to perform the visualization which combines the
        information spread across all analyses (e.g. plotting the data of each analysis on the same subplot).

        The `visualize_before_fit_combined` method is called before the model-fit begins and is used to output images
        of quantities that do not change during the fit (e.g. the data).

        When summed analysis is used, the `visualize_before_fit` method is also called for each individual analysis.
        Each individual dataset may therefore also be visualized in that function. This method is specifically for
        visualizing the combined information of all datasets.

        For your model-fitting problem this function will be overwritten with plotting functions specific to your
        problem.

        The example does not use analysis summing and therefore this function is not implemented.

        Parameters
        ----------
        analyses
            A list of the analysis classes used to perform the model-fit whose quantities are being visualized.
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        model
            The model which is fitted to the data, which may be used to customize the visualization.
        """
        pass

    @staticmethod
    def visualize_combined(
        analyses: List[af.Analysis],
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis: bool,
    ):
        """
        Multiple instances of the `Analysis` class can be summed together, meaning that the model is fitted to all
        datasets simultaneously via a summed likelihood function.

        The function receives as input a list of instances of every `Analysis` class which is being used to perform
        the summed analysis fit. This is used which is used to perform the visualization which combines the
        information spread across all analyses (e.g. plotting the data of each analysis on the same subplot).

        The `visualize_combined` method is called throughout the non-linear search and is used to output images
        indicating the quality of the fit so far.

        When summed analysis is used, the `visualize_before_fit` method is also called for each individual analysis.
        Each individual dataset may therefore also be visualized in that function. This method is specifically for
        visualizing the combined information of all datasets.

        For your model-fitting problem this function will be overwritten with plotting functions specific to your
        problem.

        The example does not use analysis summing and therefore this function is not implemented.

        Parameters
        ----------
        analyses
            A list of the analysis classes used to perform the model-fit whose quantities are being visualized.
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        model
            The model which is fitted to the data, which may be used to customize the visualization.
        """
        pass


from typing import Optional

from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.initializer import AbstractInitializer
from autofit.non_linear.search.mle.pyswarms.search.abstract import AbstractPySwarms


class PySwarmsGlobal(AbstractPySwarms):
    __identifier_fields__ = (
        "n_particles",
        "cognitive",
        "social",
        "inertia",
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            initializer: Optional[AbstractInitializer] = None,
            iterations_per_full_update: int = None,
            iterations_per_quick_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        A PySwarms Particle Swarm MLE global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms

        https://pyswarms.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating PySwarms Search")

    def search_internal_from(self, model, fitness, bounds, init_pos):
        """Get the static Dynesty sampler which performs the non-linear search, passing it all associated input Dynesty
        variables."""

        import pyswarms

        options = {
            "c1": self.config_dict_search["cognitive"],
            "c2": self.config_dict_search["social"],
            "w": self.config_dict_search["inertia"]
        }

        filter_list = ["cognitive", "social", "inertia"]
        config_dict = {key: value for key, value in self.config_dict_search.items() if key not in filter_list}

        return pyswarms.global_best.GlobalBestPSO(
            dimensions=model.prior_count,
            bounds=bounds,
            init_pos=init_pos,
            options=options,
            **config_dict
        )


from abc import ABC
import numpy as np
from typing import Optional, Dict

from autofit.mapper.prior.arithmetic.compound import CompoundPrior, Compound
from autofit.mapper.prior_model.abstract import AbstractPriorModel


class ComparisonAssertion(CompoundPrior, Compound, ABC):
    def __init__(self, lower, greater, name=""):
        super().__init__(lower, greater)
        self._name = name

    def __gt__(self, other):
        return CompoundAssertion(self, self._left > other)

    def __lt__(self, other):
        return CompoundAssertion(self, self._right < other)

    def __ge__(self, other):
        return CompoundAssertion(self, self._left >= other)

    def __le__(self, other):
        return CompoundAssertion(self, self._right <= other)


class GreaterThanLessThanAssertion(ComparisonAssertion):
    def _instance_for_arguments(self, arguments, ignore_assertions=False, xp=np):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        lower = self.left_for_arguments(
            arguments,
            ignore_assertions=ignore_assertions,
        )
        greater = self.right_for_arguments(
            arguments,
            ignore_assertions=ignore_assertions,
        )
        return lower < greater


class GreaterThanLessThanEqualAssertion(ComparisonAssertion):
    def _instance_for_arguments(
        self,
        arguments,
        ignore_assertions=False,
        xp=np,
    ):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        return self.left_for_arguments(
            arguments,
            ignore_assertions=ignore_assertions,
        ) <= self.right_for_arguments(
            arguments,
            ignore_assertions=ignore_assertions,
        )


class CompoundAssertion(AbstractPriorModel, Compound):
    def __init__(self, assertion_1, assertion_2, name=""):
        super().__init__()
        self.assertion_1 = assertion_1
        self.assertion_2 = assertion_2
        self._name = name

    def _instance_for_arguments(
        self,
        arguments,
        ignore_assertions=False,
        xp=np,
    ):
        return self.assertion_1.instance_for_arguments(
            arguments,
            ignore_assertions,
        ) and self.assertion_2.instance_for_arguments(
            arguments,
            ignore_assertions,
        )

    def dict(self) -> dict:
        return {
            "type": "compound",
            "compound_type": self.__class__.__name__,
            "assertion_1": self.assertion_1.dict(),
            "assertion_2": self.assertion_2.dict(),
        }

    @classmethod
    def from_dict(
        cls,
        d,
        reference: Optional[Dict[str, str]] = None,
        loaded_ids: Optional[dict] = None,
    ):
        return cls(
            Compound.from_dict(d["assertion_1"], reference, loaded_ids),
            Compound.from_dict(d["assertion_2"], reference, loaded_ids),
        )


def unwrap(obj):
    try:
        return obj._value
    except AttributeError:
        return obj



from typing import Union

from ..sqlalchemy_ import sa

from autofit.mapper.prior import abstract
from autofit.mapper.prior_model import prior_model
from autofit.mapper.prior_model import collection

from .model import Object


def extra_children(source):
    return [
        (f"_assertions", source.assertions),
    ]


class Collection(Object):
    """
    A collection
    """

    __tablename__ = "collection_prior_model"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "collection_prior_model"}

    @classmethod
    def _from_object(cls, source: Union[collection.Collection, list, dict]):
        instance = cls()
        if isinstance(source, collection.Collection):
            extra_children_ = extra_children(source)
            instance._add_children(extra_children_)
        else:
            source = collection.Collection(source)

        instance._add_children(source.items())

        instance.cls = collection.Collection
        return instance


class Model(Object):
    """
    A prior model
    """

    __tablename__ = "prior_model"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "prior_model"}

    @classmethod
    def _from_object(
        cls,
        model: prior_model.Model,
    ):
        instance = cls()
        instance.cls = model.cls
        instance._add_children(model.items() + extra_children(model))
        return instance

    def _make_instance(self):
        instance = object.__new__(prior_model.Model)
        instance.cls = self.cls
        instance._assertions = []

        return instance


class Prior(Object):
    """
    A prior
    """

    __tablename__ = "prior"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey("object.id"),
        primary_key=True,
        index=True,
    )

    __mapper_args__ = {"polymorphic_identity": "prior"}

    @classmethod
    def _from_object(cls, model: abstract.Prior):
        instance = cls()
        instance.cls = type(model)
        instance._add_children(
            [(key, getattr(model, key)) for key in model.__database_args__]
        )
        return instance

    def __call__(self):
        """
        Create the real instance for this object, with child
        attributes attached.

        If the instance implements __setstate__ then this is
        called with a dictionary of instantiated children.
        """
        arguments = {child.name: child() for child in self.children}
        return self.cls(**arguments)



import sys
from typing import Optional, List, Union, Callable, Type
from pathlib import Path

from PIL import Image

from autofit.aggregator.search_output import SearchOutput
from autofit.aggregator.aggregator import Aggregator

import re
from enum import Enum


def subplot_filename(subplot: Enum) -> str:
    subplot_type = subplot.__class__
    return (
        re.sub(
            r"([A-Z])",
            r"_\1",
            subplot_type.__name__,
        )
        .lower()
        .lstrip("_")
    )


class SubplotFitImage:
    def __init__(
        self,
        image: Image.Image,
        suplot_type,
    ):
        """
        The subplot_fit image associated with one fit.

        Parameters
        ----------
        image
            The subplot_fit image.
        """
        self._image = image

        x_max = 0
        y_max = 0

        for subplot in suplot_type:
            x, y = subplot.value
            x_max = max(x_max, x)
            y_max = max(y_max, y)

        self._single_image_width = self._image.width // (x_max + 1)
        self._single_image_height = self._image.height // (y_max + 1)

    def image_at_coordinates(
        self,
        x: int,
        y: int,
    ) -> Image.Image:
        """
        Extract the image at the specified coordinates.

        Parameters
        ----------
        x
        y
            The integer coordinates of the plot (see Subplot).

        Returns
        -------
        The extracted image.
        """
        return self._image.crop(
            (
                x * self._single_image_width,
                y * self._single_image_height,
                (x + 1) * self._single_image_width,
                (y + 1) * self._single_image_height,
            )
        )


class AggregateImages:
    def __init__(
        self,
        aggregator: Union[Aggregator, List[SearchOutput]],
    ):
        """
        Extracts images from the aggregator and combines them into one
        overall image.

        Parameters
        ----------
        aggregator
            The aggregator containing the fit results.
        """
        if len(aggregator) == 0:
            raise ValueError("The aggregator is empty.")

        self._aggregator = aggregator
        self._source_images = None

    def extract_image(
        self,
        subplots: List[Union[Enum, List[Image.Image], Callable]],
        subplot_width: Optional[int] = sys.maxsize,
        transpose: bool = False,
    ) -> Image.Image:
        """
        Extract the images at the specified subplots and combine them into
        one overall image including those subplots for all fits in the
        aggregator.

        Parameters
        ----------
        subplots
            The subplots to output. These can be:
            - enum values
            - lists of the same length as the aggregator output
            - functions that take a SearchOutput as an argument
        subplot_width
            Defines the width of each subplot in number of images.
            If this is greater than the number of subplots then it defaults to
            the number of subplots.
            If this is less than the number of subplots then it causes the
            images to wrap.
        transpose
            If True the output image is transposed before being returned, else it
            is returned as is.

        Returns
        -------
        The combined image.
        """
        matrix = []
        for i, result in enumerate(self._aggregator):
            matrix.extend(
                self._matrix_for_result(
                    i,
                    result,
                    subplots,
                    subplot_width=subplot_width,
                )
            )

        if transpose:

            matrix = [list(row) for row in zip(*matrix)]

        return self._matrix_to_image(matrix)

    def output_to_folder(
        self,
        folder: Path,
        name: Union[str, List[str]],
        subplots: List[Union[List[Image.Image], Callable]],
        subplot_width: Optional[int] = sys.maxsize,
    ):
        """
        Output one subplot image for each fit in the aggregator.

        Parameters
        ----------
        folder
            The target folder in which to store the subplots.
        subplots
            The subplots to output. These can be:
            - enum values
            - lists of the same length as the aggregator output
            - functions that take a SearchOutput as an argument
        subplot_width
            Defines the width of each subplot in number of images.
            If this is greater than the number of subplots then it defaults to
            the number of subplots.
            If this is less than the number of subplots then it causes the
            images to wrap.
        name
            The attribute of each fit to use as the name of the output file.
            OR a list of names, one for each fit.
        """
        if len(subplots) == 0:
            raise ValueError("At least one subplot must be provided.")

        folder.mkdir(exist_ok=True, parents=True)

        for i, result in enumerate(self._aggregator):
            image = self._matrix_to_image(
                self._matrix_for_result(
                    i,
                    result,
                    subplots,
                    subplot_width=subplot_width,
                )
            )

            if isinstance(name, str):
                output_name = getattr(result, name)
            else:
                output_name = name[i]

            output_path = folder / output_name
            output_path.parent.mkdir(parents=True, exist_ok=True)
            image.save(folder / f"{output_name}.png")

    @staticmethod
    def _matrix_for_result(
        i: int,
        result: SearchOutput,
        subplots: List[Union[List[Image.Image], Callable]],
        subplot_width: int = sys.maxsize,
    ) -> List[List[Image.Image]]:
        """
        Create a matrix of images each in the position they will be in the
        final image.

        Parameters
        ----------
        result
            The fit result.
        subplots
            The subplots to output. These can be:
            - enum values
            - lists of the same length as the aggregator output
            - functions that take a SearchOutput as an argument
        subplot_width
            The number of subplots to include in each row of the matrix.

        Returns
        -------
        The matrix of images.
        """
        _images = {}

        def get_image(subplot_: Enum) -> SubplotFitImage:
            """
            Get the image for the subplot.

            This assumes that the subplot filename is the same as the subplot
            class name but using snake_case.

            Parameters
            ----------
            subplot_
                The type of subplot to get the image for.

            Returns
            -------
            The image for the subplot.
            """
            subplot_type = subplot_.__class__
            if subplot_type not in _images:
                _images[subplot_type] = SubplotFitImage(
                    result.image(
                        subplot_filename(subplot_),
                    ),
                    subplot_type
                )
            return _images[subplot_type]

        matrix = []
        row = []
        for subplot in subplots:
            if isinstance(subplot, Enum):
                row.append(
                    get_image(subplot).image_at_coordinates(
                        *subplot.value,
                    )
                )
            elif isinstance(subplot, list):
                if not isinstance(subplot[i], Image.Image):
                    raise TypeError(
                        "The subplots must be of type Subplot or a list of "
                        "images or a function that takes a SearchOutput as an "
                        "argument."
                    )
                row.append(subplot[i])
            else:
                try:
                    row.append(subplot(result))
                except TypeError:
                    raise TypeError(
                        "The subplots must be of type Subplot or a list of "
                        "images or a function that takes a SearchOutput as an "
                        "argument."
                    )

            if len(row) == subplot_width:
                matrix.append(row)
                row = []

        if len(row) > 0:
            matrix.append(row)

        return matrix

    @staticmethod
    def _matrix_to_image(matrix: List[List[Image.Image]]) -> Image.Image:
        """
        Create an image including all the images in the matrix.

        Parameters
        ----------
        matrix
            The matrix of images to combine into one image

        Returns
        -------
        The combined image.
        """
        total_width = max(sum(image.width for image in row) for row in matrix)
        total_height = max(
            sum(image.height for image in column) for column in list(zip(*matrix))
        )

        image = Image.new("RGB", (total_width, total_height))

        y_offset = 0
        for row in matrix:
            x_offset = 0
            for subplot_image in row:
                image.paste(subplot_image, (x_offset, y_offset))
                x_offset += subplot_image.width
            y_offset += row[0].height

        return image



from autofit.non_linear.plot.samples_plotters import SamplesPlotter
from autofit.non_linear.plot.mcmc_plotters import MCMCPlotter
from autofit.non_linear.plot.mle_plotters import MLEPlotter
from autofit.non_linear.plot.nest_plotters import NestPlotter
from autofit.non_linear.plot.output import Output


import numpy as np
from typing import Dict, Optional

import autofit as af

from autofit.example.result import ResultExample
from autofit.example.visualize import VisualizerExample

"""
The `analysis.py` module contains the dataset and log likelihood function which given a model instance (set up by
the non-linear search) fits the dataset and returns the log likelihood of that model.
"""


class Analysis(af.Analysis):

    """
    This over-write means the `Visualizer` class is used for visualization throughout the model-fit.

    This `VisualizerExample` object is in the `autofit.example.visualize` module and is used to customize the
    plots output during the model-fit.

    It has been extended with visualize methods that output visuals specific to the fitting of `1D` data.
    """
    Visualizer = VisualizerExample

    """
    This over-write means the `ResultExample` class is returned after the model-fit.

    This `ResultExample` object in the `autofit.example.result` module. 
    
    It has been extended, based on the model that is input into the analysis, to include a 
    property `max_log_likelihood_model_data`, which is the model data of the best-fit model.
    """
    Result = ResultExample

    LATENT_KEYS = ["gaussian.fwhm"]

    def __init__(self, data: np.ndarray, noise_map: np.ndarray, use_jax=False):
        """
        In this example the `Analysis` object only contains the data and noise-map. It can be easily extended,
        for more complex data-sets and model fitting problems.

        Parameters
        ----------
        data
            A 1D numpy array containing the data (e.g. a noisy 1D Gaussian) fitted in the workspace examples.
        noise_map
            A 1D numpy array containing the noise values of the data, used for computing the goodness of fit
            metric.
        """
        super().__init__(use_jax=use_jax)

        self.data = data
        self.noise_map = noise_map

    def log_likelihood_function(self, instance: af.ModelInstance, xp=np) -> float:
        """
        Determine the log likelihood of a fit of multiple profiles to the dataset.

        Parameters
        ----------
        instance : af.Collection
            The model instances of the profiles.

        Returns
        -------
        The log likelihood value indicating how well this model fit the dataset.
        """
        model_data_1d = self.model_data_1d_from(instance=instance)

        residual_map = self.data - model_data_1d
        chi_squared_map = (residual_map / self.noise_map) ** 2.0
        log_likelihood = -0.5 * sum(chi_squared_map)

        return log_likelihood

    def model_data_1d_from(self, instance: af.ModelInstance) -> np.ndarray:
        """
        Returns the model data of a the 1D profiles.

        The way this is generated changes depending on if the model is a `Model` (therefore having only one profile)
        or a `Collection` (therefore having multiple profiles).

        If its a model, the model component's `model_data_from` is called and the output returned.
        For a collection, each components `model_data_from` is called, iterated through and summed
        to return the combined model data.

        Parameters
        ----------
        instance
            The model instance of the profile or collection of profiles.

        Returns
        -------
        The model data of the profiles.
        """

        xvalues = self._xp.arange(self.data.shape[0])
        model_data_1d = self._xp.zeros(self.data.shape[0])

        try:
            for profile in instance:
                try:
                    model_data_1d += profile.model_data_from(
                        xvalues=xvalues,
                        xp=self._xp
                    )
                except AttributeError:
                    pass
        except TypeError:
            model_data_1d += instance.model_data_from(xvalues=xvalues)

        return model_data_1d

    def save_attributes(self, paths: af.DirectoryPaths):
        """
        Before the model-fit via the non-linear search begins, this routine saves attributes of the `Analysis` object
        to the `files` folder such that they can be loaded after the analysis using PyAutoFit's database and
        aggregator tools.

        For this analysis the following are output:

        - The dataset's data.
        - The dataset's noise-map.

        It is common for these attributes to be loaded by many of the template aggregator functions given in the
        `aggregator` modules. For example, when using the database tools to reperform a fit, this will by default
        load the dataset, settings and other attributes necessary to perform a fit using the attributes output by
        this function.

        Parameters
        ----------
        paths
            The paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        """
        paths.save_json(name="data", object_dict=self.data.tolist(), prefix="dataset")
        paths.save_json(
            name="noise_map", object_dict=self.noise_map.tolist(), prefix="dataset"
        )

    def make_result(
        self,
        samples_summary: af.SamplesSummary,
        paths: af.AbstractPaths,
        samples: Optional[af.SamplesPDF] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[object] = None,
    ) -> Result:
        """
        Returns the `Result` of the non-linear search after it is completed.

        The result type is defined as a class variable in the `Analysis` class (see top of code under the python code
        `class Analysis(af.Analysis)`.

        The result can be manually overwritten by a user to return a user-defined result object, which can be extended
        with additional methods and attribute specific to the model-fit.

        This example class does example this, whereby the analysis result has been over written with the `ResultExample`
        class, which contains a property `max_log_likelihood_model_data_1d` that returns the model data of the
        best-fit model. This API means you can customize your result object to include whatever attributes you want
        and therefore make a result object specific to your model-fit and model-fitting problem.

        The `Result` object you return can be customized to include:

        - The samples summary, which contains the maximum log likelihood instance and median PDF model.

        - The paths of the search, which are used for loading the samples and search internal below when a search
        is resumed.

        - The samples of the non-linear search (e.g. MCMC chains) also stored in `samples.csv`.

        - The non-linear search used for the fit in its internal representation, which is used for resuming a search
        and making bespoke visualization using the search's internal results.

        - The analysis used to fit the model (default disabled to save memory, but option may be useful for certain
        projects).

        Parameters
        ----------
        samples_summary
            The summary of the samples of the non-linear search, which include the maximum log likelihood instance and
            median PDF model.
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        samples
            The samples of the non-linear search, for example the chains of an MCMC run.
        search_internal
            The internal representation of the non-linear search used to perform the model-fit.
        analysis
            The analysis used to fit the model.

        Returns
        -------
        Result
            The result of the non-linear search, which is defined as a class variable in the `Analysis` class.
        """
        return self.Result(
            samples_summary=samples_summary,
            paths=paths,
            samples=samples,
            search_internal=search_internal,
            analysis=self,
        )

    def compute_latent_variables(self, parameters, model) -> Dict[str, float]:
        """
        A latent variable is not a model parameter but can be derived from the model. Its value and errors may be
        of interest and aid in the interpretation of a model-fit.

        For example, for the simple 1D Gaussian example, it could be the full-width half maximum (FWHM) of the
        Gaussian. This is not included in the model but can be easily derived from the Gaussian's sigma value.

        By overwriting this method we can manually specify latent variables that are calculated and output to
        a `latent.csv` file, which mirrors the `samples.csv` file.

        In the example below, the `latent.csv` file will contain one column with the FWHM of every Gausian model
        sampled by the non-linear search.

        This function is called at the end of search, following one of two schemes depending on the settings in
        `output.yaml`:

        1) Call for every search sample, which produces a complete `latent/samples.csv` which mirrors the normal
        `samples.csv` file but takes a long time to compute.

        2) Call only for N random draws from the posterior inferred at the end of the search, which only produces a
        `latent/latent_summary.json` file with the median and 1 and 3 sigma errors of the latent variables but is
        fast to compute.

        Parameters
        ----------
        instance
            The instances of the model which the latent variable is derived from.

        Returns
        -------

        """
        instance = model.instance_from_vector(vector=parameters)

        try:
            return (instance.fwhm, )
        except AttributeError:
            try:
                return (instance[0].fwhm,)
            except AttributeError:
                return (instance[0].gaussian.fwhm,)